from touchstone.lib import exceptions
from touchstone.lib.configurers.i_configurable import IConfigurable
from touchstone.lib.health_checks.http_health_check import HttpHealthCheck
from touchstone.lib.managers.docker_manager import DockerManager
from touchstone.lib.networking.docker_network import DockerNetwork
from touchstone.lib.networking.i_network import INetwork
from touchstone.lib.nodes.mocks.behaviors.i_behavior import IBehavior
from touchstone.lib.nodes.mocks.behaviors.i_http_behavior import IHttpBehavior, IHttpSetup, IHttpVerify
from touchstone.lib.nodes.mocks.docker.http.docker_http_setup import DockerHttpSetup
from touchstone.lib.nodes.mocks.docker.http.docker_http_verify import DockerHttpVerify
from touchstone.lib.nodes.mocks.docker.i_runnable_docker import IRunnableDocker


class DockerHttp(IRunnableDocker, IHttpBehavior):
    def __init__(self, defaults_configurer: IConfigurable, health_check: HttpHealthCheck, setup: DockerHttpSetup,
                 verify: DockerHttpVerify, docker_manager: DockerManager, docker_network: DockerNetwork):
        self.__defaults_configurer = defaults_configurer
        self.__health_check = health_check
        self.__setup = setup
        self.__verify = verify
        self.__docker_manager = docker_manager
        self.__docker_network = docker_network

    def get_behavior(self) -> IBehavior:
        return self

    def get_network(self) -> INetwork:
        return self.__docker_network

    def initialize(self):
        self.__setup.set_url(self.__docker_network.external_url())
        self.__verify.set_url(self.__docker_network.external_url())
        self.__setup.init(self.__defaults_configurer.get_config())

    def start(self):
        run_result = self.__docker_manager.run_background_image('holomekc/wiremock-gui:2.25.1', port=8080,
                                                                exposed_port=9090)
        self.__docker_network.set_container_id(run_result.container_id)
        self.__docker_network.set_internal_port(run_result.internal_port)
        self.__docker_network.set_external_port(run_result.external_port)
        self.__docker_network.set_ui_port(run_result.external_port)
        self.__docker_network.set_ui_endpoint('/__admin/webapp')
        self.__docker_network.set_prefix('http://')

    def stop(self):
        if self.__docker_network.container_id():
            self.__docker_manager.stop_container(self.__docker_network.container_id())

    def reset(self):
        self.__setup.init(self.__defaults_configurer.get_config())

    def is_healthy(self) -> bool:
        self.__health_check.set_url(self.__docker_network.ui_url())
        return self.__health_check.is_healthy()

    def setup(self) -> IHttpSetup:
        if not self.__setup:
            raise exceptions.MockException('Setup unavailable. Mock is still starting.')
        return self.__setup

    def verify(self) -> IHttpVerify:
        if not self.__verify:
            raise exceptions.MockException('Verify unavailable. Mock is still starting.')
        return self.__verify

    def url(self) -> str:
        return self.__docker_network.external_url()



class TouchstoneException(Exception):
    pass


class ServiceException(Exception):
    pass


class TestException(Exception):
    pass


class MockException(Exception):
    pass


class MockNotSupportedException(Exception):
    pass


class ContainerException(Exception):
    pass



from touchstone.lib import exceptions
from touchstone.lib.configurers.i_configurable import IConfigurable
from touchstone.lib.nodes.mocks.behaviors.i_behavior import IBehavior
from touchstone.lib.nodes.mocks.behaviors.i_filesystem_behavior import IFilesystemBehavior, IFilesystemVerify, \
    IFilesystemSetup
from touchstone.lib.nodes.mocks.local.filesystem.local_filesystem_setup import LocalFilesystemSetup
from touchstone.lib.nodes.mocks.local.filesystem.local_filesystem_verify import LocalFilesystemVerify
from touchstone.lib.nodes.mocks.local.i_runnable_local import IRunnableLocal


class LocalFilesystem(IRunnableLocal, IFilesystemBehavior):
    def __init__(self, defaults_configurer: IConfigurable, files_path: str, setup: LocalFilesystemSetup,
                 verify: LocalFilesystemVerify):
        self.__defaults_configurer = defaults_configurer
        self.__files_path = files_path
        self.__setup = setup
        self.__verify = verify

    def get_behavior(self) -> IBehavior:
        return self

    def initialize(self):
        pass

    def start(self):
        self.__setup.reset()

    def stop(self):
        self.__setup.delete_defaults()

    def reset(self):
        self.__setup.reset()

    def is_healthy(self) -> bool:
        return True

    def setup(self) -> IFilesystemSetup:
        if not self.__setup:
            raise exceptions.MockException('Setup unavailable. Mock is still starting.')
        return self.__setup

    def verify(self) -> IFilesystemVerify:
        if not self.__verify:
            raise exceptions.MockException('Verify unavailable. Mock is still starting.')
        return self.__verify

    def get_io_path(self) -> str:
        return self.__files_path



import yaml

from touchstone import common
from touchstone.lib.configurers.i_configurable import IConfigurable


class FileConfigurer(IConfigurable):
    def __init__(self, config_path: str = None):
        self.__config_path = config_path
        self.__override_config = {}

    def get_config(self) -> dict:
        if not self.__config_path:
            return {}
        with open(self.__config_path, 'r') as file:
            config = yaml.safe_load(file)
        return common.dict_merge(config, self.__override_config)

    def merge_config(self, other: dict):
        self.__override_config = other



import abc
from typing import List

from touchstone.lib.nodes.mocks.behaviors.i_behavior import IBehavior


class IMysqlSetup(object):
    @abc.abstractmethod
    def execute(self, database: str, sql: str):
        """Executes arbitrary SQL on the given database."""
        pass

    @abc.abstractmethod
    def insert_row(self, database: str, table: str, data: dict):
        """Inserts a dictionary of key-value pairs into the given database and table. If the config option,
        "camel_to_snake" is set (default True), the dictionary keys will be converted from camel case to
        snake case."""
        pass

    @abc.abstractmethod
    def insert_rows(self, database: str, table: str, data: List[dict]):
        """Inserts a list of dictionaries of key-value pairs into the given database and table. If the config option,
        "camel_to_snake" is set (default True), the dictionary keys will be converted from camel case to
        snake case."""
        pass


class IMysqlVerify(object):
    @abc.abstractmethod
    def row_exists(self, database: str, table: str, where_conditions: dict, num_expected: int = 1) -> bool:
        """Returns True if the given where conditions are found in the given database. If num_expected is set to None,
        any number of rows will be considered passing."""
        pass

    @abc.abstractmethod
    def row_does_not_exist(self, database: str, table: str, where_conditions: dict) -> bool:
        """Returns True if the given where conditions are not found in the given database."""
        pass


class IMysqlBehavior(IBehavior):
    DEFAULT_CONFIG = {
        'camel_to_snake': True,
        'snapshot_databases': False
    }

    @abc.abstractmethod
    def setup(self) -> IMysqlSetup:
        pass

    @abc.abstractmethod
    def verify(self) -> IMysqlVerify:
        pass



import abc
from typing import List

from touchstone.lib.nodes.mocks.behaviors.i_behavior import IBehavior


class IMongodbSetup(object):
    def command(self, database: str, command: dict):
        """Execute an arbitrary command on the database."""
        pass

    def insert_document(self, database: str, collection: str, document: dict):
        """Inserts a document into the given database and collection."""
        pass

    def insert_documents(self, database: str, collection: str, documents: List[dict]):
        """Inserts multiple documents into the given database and collection."""
        pass


class IMongodbVerify(object):
    def document_exists(self, database: str, collection: str, document: dict, num_expected: int = 1) -> bool:
        """Returns True if a document exists in the given database and collection. If num_expected is set to None,
        any number of documents will be considered passing."""
        pass


class IMongodbBehavior(IBehavior):
    @abc.abstractmethod
    def setup(self) -> IMongodbSetup:
        pass

    @abc.abstractmethod
    def verify(self) -> IMongodbVerify:
        pass



from typing import List, Tuple

from touchstone.lib import exceptions
from touchstone.lib.nodes.services.i_executable import IExecutable
from touchstone.lib.nodes.services.i_runnable import IRunnable
from touchstone.lib.nodes.services.i_service import IService
from touchstone.lib.nodes.services.i_service_executor import IServiceExecutor
from touchstone.lib.nodes.services.i_testable import ITestable


class Services(IServiceExecutor):
    def __init__(self):
        self.__services: List[IService] = []
        self.__services_running = False
        self.__environment_vars: List[Tuple[str, str]] = []

    def add_service(self, service: IService):
        self.__services.append(service)

    def add_environment_vars(self, environment_vars: List[Tuple[str, str]] = []):
        self.__environment_vars.extend(environment_vars)

    def start(self):
        if self.__services_running:
            print('Services have already been started. They cannot be started again.')
        else:
            runnables: List[IRunnable] = []
            for service in self.__services:
                if isinstance(service, IRunnable):
                    runnables.append(service)
            print(f'Starting services {[_.get_name() for _ in runnables]}...')
            for service in runnables:
                service.start(self.__environment_vars)
            self.__services_running = True
            for service in runnables:
                service.wait_until_healthy()
            print('Finished starting services.\n')

    def stop(self):
        print('Stopping services...')
        for service in self.__services:
            if isinstance(service, IRunnable):
                service.stop()
        self.__services_running = False

    def execute(self, service_name: str):
        found = False
        for service in self.__services:
            if service.get_name() == service_name and isinstance(service, IExecutable):
                found = True
                service.execute(self.__environment_vars)
        if not found:
            raise exceptions.ServiceException(f'Service could not be found with name: "{service_name}".')

    def run_test(self, service_name: str, file_name: str, test_name: str) -> bool:
        found_service = None
        for service in self.__services:
            if service.get_name().replace(' ', '-').lower() == service_name \
                    and isinstance(service, ITestable):
                found_service = service
        if not found_service:
            print(f'No testable service could be found with the name "{service_name}".')
            return False

        return found_service.run_test(file_name, test_name)

    def run_all_tests(self) -> bool:
        for service in self.__services:
            if isinstance(service, ITestable):
                did_pass = service.run_all_tests()
                if not did_pass:
                    return False
        return True

    def are_running(self) -> bool:
        return self.__services_running



import time
import traceback
from typing import List

from touchstone.lib.nodes.mocks.mocks import Mocks
from touchstone.lib.nodes.services.i_service_executor import IServiceExecutor
from touchstone.lib.touchstone_test import TouchstoneTest


class TestClass(object):
    def __init__(self, name, clazz):
        self.name = name
        self.clazz = clazz


class TestContainer(object):
    def __init__(self, file):
        self.file = file
        self.test_classes: List[TestClass] = []

    def add_test_class(self, test_class: TestClass):
        self.test_classes.append(test_class)

    def execute(self, service_url: str, mocks: Mocks, service_executor: IServiceExecutor) -> bool:
        all_passed = True
        for test_class in self.test_classes:
            class_instance: TouchstoneTest = test_class.clazz(service_url, mocks, service_executor)
            test = Test(test_class.name, class_instance)
            print(f'{test.name} :: RUNNING')
            mocks.reset()

            did_pass = False
            try:
                did_pass = test.run()
            except Exception:
                traceback.print_exc()
                all_passed = False

            if not did_pass:
                print(f'{test.name} :: FAILED\n')
                all_passed = False
            else:
                print(f'{test.name} :: PASSED\n')
        return all_passed


class Test(object):
    def __init__(self, name: str, touchstone_test: TouchstoneTest):
        self.name = name
        self.__touchstone_test = touchstone_test

    def run(self) -> bool:
        given = self.__touchstone_test.given()
        result = self.__touchstone_test.when(given)
        time.sleep(self.__touchstone_test.processing_period())
        did_pass = self.__touchstone_test.then(given, result)
        return did_pass



import logging
import os
import re
from typing import List

import yaml

from touchstone import __version__
from touchstone.lib import exceptions

logger = logging.getLogger('touchstone')
__camel_to_snake_pattern = re.compile(r'(?<!^)(?=[A-Z])')


def sanity_check_passes() -> bool:
    # Ensure paths and files exist
    touchstone_path = os.path.join(os.getcwd(), 'touchstone')
    touchstone_yml_path = os.path.join(touchstone_path, 'touchstone.yml')
    defaults_path = os.path.join(touchstone_path, 'defaults')
    paths_exist = os.path.exists(touchstone_path) \
                  and os.path.exists(touchstone_yml_path) \
                  and os.path.exists(defaults_path)
    if not paths_exist:
        print('The touchstone directory, touchstone.yml, or the defaults directory could not be found. '
              'If touchstone has not been initialized, run \'touchstone init\'.')
        return False

    # Ensure versions are compatible
    versions_match = False
    with open(touchstone_yml_path, 'r') as file:
        touchstone_config = yaml.safe_load(file)
        given_version = touchstone_config.get('touchstone_version')
        if given_version:
            split_given_version = str.split(given_version, '.')
            split_current_version = str.split(__version__, '.')
            versions_match = split_given_version[0] == split_current_version[0] \
                             and split_given_version[1] <= split_current_version[1]
            if not versions_match:
                print(f'Version defined in touchstone.yml: "{given_version}" is not compatible with system Touchstone '
                      f'version:"{__version__}"')
        else:
            print('A touchstone version number must be defined in "touchstone.yml".')
    return versions_match


def dict_merge(base: dict, override: dict) -> dict:
    if override is None:
        return base
    return dict(list(base.items()) + list(override.items()))


def __camel_to_snake(input: str) -> str:
    return __camel_to_snake_pattern.sub('_', input).lower()


def __dict_keys_to_snake(input: dict) -> dict:
    new = {}
    for key, value in input.items():
        new[__camel_to_snake(key)] = value
    return new


def __list_dict_keys_to_snake(input: List[dict]) -> List[dict]:
    new = []
    for item in input:
        new.append(__dict_keys_to_snake(item))
    return new


def to_snake(input) -> object:
    if isinstance(input, str):
        return __camel_to_snake(input)
    if isinstance(input, dict):
        return __dict_keys_to_snake(input)
    if isinstance(input, list):
        return __list_dict_keys_to_snake(input)
    raise exceptions.TouchstoneException('Given input can not be converted to snake case.')



import urllib.error
import urllib.request

from touchstone import common
from touchstone.lib import exceptions
from touchstone.lib.health_checks.i_health_checkable import IHealthCheckable


class HttpHealthCheck(IHealthCheckable):
    def __init__(self):
        self.__url = None

    def set_url(self, url: str):
        self.__url = url

    def is_healthy(self) -> bool:
        if not self.__url:
            raise exceptions.MockException('URL must be set before checking health.')
        try:
            common.logger.debug(f'Executing HTTP health check for: {self.__url}')
            code = urllib.request.urlopen(self.__url).getcode()
            passed = code % 200 < 100
            if passed:
                common.logger.debug(f'HTTP health check passed for: {self.__url}')
            else:
                common.logger.debug(f'HTTP health check failed for: {self.__url}')
            return passed
        except (urllib.error.URLError, ConnectionResetError):
            return False



import subprocess

import pymysql

from touchstone.lib import exceptions
from touchstone.lib.configurers.i_configurable import IConfigurable
from touchstone.lib.listeners.i_services_available_listener import IServicesAvailableListener
from touchstone.lib.managers.docker_manager import DockerManager
from touchstone.lib.networking.docker_network import DockerNetwork
from touchstone.lib.networking.i_network import INetwork
from touchstone.lib.nodes.mocks.behaviors.i_behavior import IBehavior
from touchstone.lib.nodes.mocks.behaviors.i_mysql_behabior import IMysqlBehavior, IMysqlVerify, IMysqlSetup
from touchstone.lib.nodes.mocks.docker.i_runnable_docker import IRunnableDocker
from touchstone.lib.nodes.mocks.docker.mysql.docker_mysql_context import DockerMysqlContext
from touchstone.lib.nodes.mocks.docker.mysql.docker_mysql_setup import DockerMysqlSetup
from touchstone.lib.nodes.mocks.docker.mysql.docker_mysql_verify import DockerMysqlVerify
from touchstone.lib.ts_context import TsContext


class DockerMysql(IRunnableDocker, IMysqlBehavior, IServicesAvailableListener):
    __USERNAME = 'root'
    __PASSWORD = 'root'

    def __init__(self, ts_context: TsContext, defaults_configurer: IConfigurable, mysql_context: DockerMysqlContext,
                 is_dev_mode: bool, configurer: IConfigurable, setup: DockerMysqlSetup, verify: DockerMysqlVerify,
                 docker_manager: DockerManager, docker_network: DockerNetwork):
        ts_context.register_services_available_listener(self)
        self.__defaults_configurer = defaults_configurer
        self.__mysql_context = mysql_context
        self.__is_dev_mode = is_dev_mode
        self.__configurer = configurer
        self.__setup = setup
        self.__verify = verify
        self.__docker_manager = docker_manager
        self.__docker_network = docker_network
        self.__ui_container_id = None

    def get_behavior(self) -> IBehavior:
        return self

    def get_network(self) -> INetwork:
        return self.__docker_network

    def initialize(self):
        connection = pymysql.connect(host=self.__docker_network.external_host(),
                                     port=self.__docker_network.external_port(),
                                     user=self.__USERNAME,
                                     password=self.__PASSWORD,
                                     charset='utf8mb4',
                                     autocommit=True,
                                     cursorclass=pymysql.cursors.DictCursor)
        cursor = connection.cursor()
        convert_camel_to_snake = self.__configurer.get_config()['camel_to_snake']
        self.__setup.set_cursor(cursor)
        self.__setup.set_convert_camel_to_snake(convert_camel_to_snake)
        self.__verify.set_cursor(cursor)
        self.__verify.set_convert_camel_to_snake(convert_camel_to_snake)
        self.__setup.init(self.__defaults_configurer.get_config())

    def start(self):
        run_result = self.__docker_manager.run_background_image(
            'mysql:8.0.20 --default-authentication-plugin=mysql_native_password', port=3306,
            environment_vars=[('MYSQL_ROOT_PASSWORD', self.__USERNAME)])
        self.__docker_network.set_container_id(run_result.container_id)
        if self.__is_dev_mode:
            ui_run_result = self.__docker_manager.run_background_image('adminer:4.7.5-standalone',
                                                                       ui_port=8080,
                                                                       environment_vars=[
                                                                           ('ADMINER_DEFAULT_SERVER',
                                                                            self.__docker_network.internal_host())])
            self.__ui_container_id = ui_run_result.container_id
            self.__docker_network.set_ui_port(ui_run_result.ui_port)
        self.__docker_network.set_internal_port(run_result.internal_port)
        self.__docker_network.set_external_port(run_result.external_port)
        self.__docker_network.set_username(self.__USERNAME)
        self.__docker_network.set_password(self.__PASSWORD)

    def stop(self):
        if self.__docker_network.container_id():
            self.__docker_manager.stop_container(self.__docker_network.container_id())
        if self.__ui_container_id:
            self.__docker_manager.stop_container(self.__ui_container_id)

    def reset(self):
        if self.__configurer.get_config()['snapshot_databases']:
            self.__setup.recreate_databases()
            for database in self.__mysql_context.databases:
                subprocess.run(f'docker exec {self.__docker_network.container_id()} sh -c "mysql -u {self.__USERNAME} '
                               f'-p{self.__PASSWORD} {database} < dump-{database}.sql"', shell=True,
                               stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        else:
            self.__setup.init(self.__defaults_configurer.get_config())

    def is_healthy(self) -> bool:
        try:
            pymysql.connect(host=self.__docker_network.external_host(),
                            port=self.__docker_network.external_port(),
                            user=self.__USERNAME,
                            password=self.__PASSWORD)
            return True
        except Exception:
            return False

    def setup(self) -> IMysqlSetup:
        if not self.__setup:
            raise exceptions.MockException('Setup unavailable. Mock is still starting.')
        return self.__setup

    def verify(self) -> IMysqlVerify:
        if not self.__verify:
            raise exceptions.MockException('Verify unavailable. Mock is still starting.')
        return self.__verify

    def services_available(self):
        if self.__configurer.get_config()['snapshot_databases']:
            for database in self.__mysql_context.databases:
                subprocess.run(
                    f'docker exec {self.__docker_network.container_id()} sh -c "mysqldump -u {self.__USERNAME} '
                    f'-p{self.__PASSWORD} --no-create-db {database} > dump-{database}.sql"', shell=True,
                    stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)



import json
import os
from io import BytesIO

from minio import Minio

from touchstone.lib.nodes.mocks.behaviors.i_s3_behavior import IS3Setup


class DockerS3Setup(IS3Setup):
    def __init__(self):
        self.__s3_client = None

    def set_s3_client(self, s3_client: Minio):
        self.__s3_client = s3_client

    def init(self, path: str, defaults: dict):
        for bucket in self.__s3_client.list_buckets():
            objects = self.__s3_client.list_objects(bucket.name, recursive=True)
            for m_object in objects:
                self.__s3_client.remove_object(bucket.name, m_object.object_name)
            self.__s3_client.remove_bucket(bucket.name)

        for bucket in defaults.get('buckets', []):
            bucket_name = bucket['name']
            self.__s3_client.make_bucket(bucket_name)
            for m_object in bucket.get('objects', []):
                object_path = os.path.join(path, m_object['path'])
                file_stat = os.stat(object_path)
                with open(object_path, 'rb') as data:
                    self.__s3_client.put_object(
                        bucket_name,
                        m_object['name'],
                        data,
                        file_stat.st_size,
                        m_object['content-type']
                    )

    def create_bucket(self, name: str):
        self.__s3_client.make_bucket(name)

    def put_object(self, bucket_name: str, object_name: str, data: bytes,
                   content_type: str = 'application/octet-stream'):
        length = len(data)
        data = BytesIO(data)
        self.__s3_client.put_object(bucket_name, object_name, data, length, content_type)

    def put_json_object(self, bucket_name: str, object_name: str, input_json: dict):
        input_bytes = bytes(json.dumps(input_json), encoding='utf-8')
        self.put_object(bucket_name, object_name, input_bytes, 'application/json')



import abc


class ITestable(object):
    @abc.abstractmethod
    def run_test(self, file_name, test_name) -> bool:
        pass

    @abc.abstractmethod
    def run_all_tests(self) -> bool:
        pass



import json
import urllib
import urllib.error
import urllib.parse
import urllib.request

from touchstone.lib import exceptions
from touchstone.lib.nodes.mocks.behaviors.i_http_behavior import IHttpSetup


class DockerHttpSetup(IHttpSetup):
    def __init__(self):
        super().__init__()
        self.__url = None
        self.mock_ids: list = []

    def set_url(self, url: str):
        self.__url = url

    def init(self, defaults: dict):
        # Remove all mocked endpoints
        for mock_id in self.mock_ids:
            request = urllib.request.Request(f'{self.__url}/__admin/mappings/{mock_id}', method='DELETE')
            try:
                urllib.request.urlopen(request)
            except urllib.error.HTTPError:
                pass
        self.mock_ids = []

        # Reset requests journal
        request = urllib.request.Request(f'{self.__url}/__admin/requests', method='DELETE')
        urllib.request.urlopen(request)

        for request in defaults.get('requests', []):
            self.__submit_mock(request)

    def get(self, endpoint: str, response: str, response_status: int = 200, url_pattern: bool = False,
            response_headers: dict = {'Content-Type': 'application/json'}):
        self.__check_mock_response_type(response)
        mock = {
            'request': {
                'method': 'GET',
                self.__get_url_type(url_pattern): endpoint
            },
            'response': {
                'status': response_status,
                'headers': response_headers,
                'body': response
            }
        }
        self.__submit_mock(mock)

    def post(self, endpoint: str, response: str, response_status: int = 200, url_pattern: bool = False,
             response_headers: dict = {'Content-Type': 'application/json'}):
        self.__check_mock_response_type(response)
        mock = {
            'request': {
                'method': 'POST',
                self.__get_url_type(url_pattern): endpoint
            },
            'response': {
                'status': response_status,
                'headers': response_headers,
                'body': response
            }
        }
        self.__submit_mock(mock)

    def put(self, endpoint: str, response: str, response_status: int = 200, url_pattern: bool = False,
            response_headers: dict = {'Content-Type': 'application/json'}):
        self.__check_mock_response_type(response)
        mock = {
            'request': {
                'method': 'PUT',
                self.__get_url_type(url_pattern): endpoint
            },
            'response': {
                'status': response_status,
                'headers': response_headers,
                'body': response
            }
        }
        self.__submit_mock(mock)

    def delete(self, endpoint: str, response: str, response_status: int = 200, url_pattern: bool = False,
               response_headers: dict = {'Content-Type': 'application/json'}):
        self.__check_mock_response_type(response)
        mock = {
            'request': {
                'method': 'DELETE',
                self.__get_url_type(url_pattern): endpoint
            },
            'response': {
                'status': response_status,
                'headers': response_headers,
                'body': response
            }
        }
        self.__submit_mock(mock)

    def __get_url_type(self, url_pattern: bool):
        if url_pattern:
            return 'urlPattern'
        return 'url'

    def __check_mock_response_type(self, response):
        if type(response) is not str:
            raise exceptions.MockException('Response must be of type str.')

    def __submit_mock(self, mock: dict):
        data = json.dumps(mock).encode('utf-8')
        request = urllib.request.Request(
            f'{self.__url}/__admin/mappings', data=data, headers={'Content-Type': 'application/json'})
        response = urllib.request.urlopen(request).read()
        self.mock_ids.append(json.loads(response)['id'])



from typing import List, Tuple

from touchstone.lib import exceptions
from touchstone.lib.health_checks.blocking_health_check import BlockingHealthCheck
from touchstone.lib.nodes.mocks.behaviors.i_filesystem_behavior import IFilesystemBehavior
from touchstone.lib.nodes.mocks.behaviors.i_http_behavior import IHttpBehavior
from touchstone.lib.nodes.mocks.behaviors.i_mongodb_behavior import IMongodbBehavior
from touchstone.lib.nodes.mocks.behaviors.i_mysql_behabior import IMysqlBehavior
from touchstone.lib.nodes.mocks.behaviors.i_rabbitmq_behavior import IRabbitmqBehavior
from touchstone.lib.nodes.mocks.behaviors.i_redis_behavior import IRedisBehavior
from touchstone.lib.nodes.mocks.behaviors.i_s3_behavior import IS3Behavior
from touchstone.lib.nodes.mocks.docker.runnable_docker_mock import RunnableDockerMock
from touchstone.lib.nodes.mocks.i_mockable import IMockable
from touchstone.lib.nodes.mocks.local.runnable_local_mock import RunnableLocalMock


class Mocks(object):
    def __init__(self):
        self.http: IHttpBehavior = None
        self.rabbitmq: IRabbitmqBehavior = None
        self.mongodb: IMongodbBehavior = None
        self.mysql: IMysqlBehavior = None
        self.s3: IS3Behavior = None
        self.filesystem: IFilesystemBehavior = None
        self.redis: IRedisBehavior = None
        self.__runnable_docker_mocks: List[RunnableDockerMock] = []
        self.__runnable_local_mocks: List[RunnableLocalMock] = []
        self.__mocks_running = False

    def register_mock(self, mock: IMockable):
        if isinstance(mock, RunnableLocalMock):
            self.__runnable_local_mocks.append(mock)
        elif isinstance(mock, RunnableDockerMock):
            self.__runnable_docker_mocks.append(mock)
        else:
            raise exceptions.MockException(f'Unsupported mock type: {mock}')

    def start(self):
        if self.__mocks_running:
            print('Mocks have already been started. They cannot be started again.')
        else:
            if self.__runnable_docker_mocks:
                print(f'Starting Docker mocks {[_.get_pretty_name() for _ in self.__runnable_docker_mocks]}...')
                for mock in self.__runnable_docker_mocks:
                    mock.start()
            if self.__runnable_local_mocks:
                print(f'Starting Local mocks {[_.get_pretty_name() for _ in self.__runnable_local_mocks]}...')
                for mock in self.__runnable_local_mocks:
                    mock.start()
            print('Waiting for mocks to become healthy...')
            self.__wait_for_healthy_mocks()
            self.__mocks_running = True
            print('Finished starting mocks.\n')

    def stop(self):
        print('Stopping mocks...')
        for mock in self.__runnable_docker_mocks:
            mock.stop()
        for mock in self.__runnable_local_mocks:
            mock.stop()
        self.__mocks_running = False

    def are_running(self):
        return self.__mocks_running

    def reset(self):
        for mock in self.__runnable_docker_mocks:
            mock.reset()
        for mock in self.__runnable_local_mocks:
            mock.reset()

    def environment_vars(self) -> List[Tuple[str, str]]:
        envs = []
        for mock in self.__runnable_docker_mocks:
            name = mock.get_name().upper()
            envs.append((f'TS_{name}_HOST', mock.get_network().internal_host()))
            envs.append((f'TS_{name}_PORT', mock.get_network().internal_port()))
            envs.append((f'TS_{name}_URL', mock.get_network().internal_url()))
            envs.append((f'TS_{name}_USERNAME', mock.get_network().username()))
            envs.append((f'TS_{name}_PASSWORD', mock.get_network().password()))
        return envs

    def print_available_mocks(self):
        for mock in self.__runnable_docker_mocks:
            message = f'Mock {mock.get_pretty_name()} running with UI: {mock.get_network().ui_url()}'
            if mock.get_network().username():
                message += f' and Username: "{mock.get_network().username()}", ' \
                           f'Password: "{mock.get_network().password()}"'
            print(message)
        for mock in self.__runnable_local_mocks:
            print(f'Mock {mock.get_pretty_name()} running')

    def __wait_for_healthy_mocks(self):
        for mock in self.__runnable_docker_mocks:
            blocking_health_check = BlockingHealthCheck(5, 10, mock)
            is_healthy = blocking_health_check.wait_until_healthy()
            if not is_healthy:
                raise exceptions.MockException(
                    f'Mock {mock.get_pretty_name()} never became healthy and timed out on initialization.')
        for mock in self.__runnable_local_mocks:
            blocking_health_check = BlockingHealthCheck(5, 10, mock)
            is_healthy = blocking_health_check.wait_until_healthy()
            if not is_healthy:
                raise exceptions.MockException(
                    f'Mock {mock.get_pretty_name()} never became healthy and timed out on initialization.')



import os
from typing import Optional

from touchstone.lib.configurers.BasicConfigurer import BasicConfigurer
from touchstone.lib.configurers.FileConfigurer import FileConfigurer
from touchstone.lib.health_checks.http_health_check import HttpHealthCheck
from touchstone.lib.managers.docker_manager import DockerManager
from touchstone.lib.networking.docker_network import DockerNetwork
from touchstone.lib.nodes.mocks.behaviors.i_mysql_behabior import IMysqlBehavior
from touchstone.lib.nodes.mocks.behaviors.i_rabbitmq_behavior import IRabbitmqBehavior
from touchstone.lib.nodes.mocks.docker.http.docker_http import DockerHttp
from touchstone.lib.nodes.mocks.docker.http.docker_http_setup import DockerHttpSetup
from touchstone.lib.nodes.mocks.docker.http.docker_http_verify import DockerHttpVerify
from touchstone.lib.nodes.mocks.docker.mongodb.docker_mongo_context import DockerMongoContext
from touchstone.lib.nodes.mocks.docker.mongodb.docker_mongodb import DockerMongodb
from touchstone.lib.nodes.mocks.docker.mongodb.docker_mongodb_setup import DockerMongodbSetup
from touchstone.lib.nodes.mocks.docker.mongodb.docker_mongodb_verify import DockerMongodbVerify
from touchstone.lib.nodes.mocks.docker.mysql.docker_mysql import DockerMysql
from touchstone.lib.nodes.mocks.docker.mysql.docker_mysql_context import DockerMysqlContext
from touchstone.lib.nodes.mocks.docker.mysql.docker_mysql_setup import DockerMysqlSetup
from touchstone.lib.nodes.mocks.docker.mysql.docker_mysql_verify import DockerMysqlVerify
from touchstone.lib.nodes.mocks.docker.rabbitmq.docker_rabbitmq import DockerRabbitmq
from touchstone.lib.nodes.mocks.docker.rabbitmq.docker_rabbitmq_context import DockerRabbitmqContext
from touchstone.lib.nodes.mocks.docker.rabbitmq.docker_rabbitmq_setup import DockerRabbitmqSetup
from touchstone.lib.nodes.mocks.docker.rabbitmq.docker_rabbitmq_verify import DockerRabbitmqVerify
from touchstone.lib.nodes.mocks.docker.redis.docker_redis import DockerRedis
from touchstone.lib.nodes.mocks.docker.redis.docker_redis_setup import DockerRedisSetup
from touchstone.lib.nodes.mocks.docker.redis.docker_redis_verify import DockerRedisVerify
from touchstone.lib.nodes.mocks.docker.runnable_docker_mock import RunnableDockerMock
from touchstone.lib.nodes.mocks.docker.s3.docker_s3 import DockerS3
from touchstone.lib.nodes.mocks.docker.s3.docker_s3_setup import DockerS3Setup
from touchstone.lib.nodes.mocks.docker.s3.docker_s3_verify import DockerS3Verify
from touchstone.lib.nodes.mocks.i_mockable import IMockable
from touchstone.lib.nodes.mocks.local.filesystem.local_filesystem import LocalFilesystem
from touchstone.lib.nodes.mocks.local.filesystem.local_filesystem_setup import LocalFilesystemSetup
from touchstone.lib.nodes.mocks.local.filesystem.local_filesystem_verify import LocalFilesystemVerify
from touchstone.lib.nodes.mocks.local.runnable_local_mock import RunnableLocalMock
from touchstone.lib.ts_context import TsContext


class MockFactory(object):
    def __init__(self, ts_context: TsContext, is_dev_mode: bool, root: str, defaults_paths: dict, configs: dict,
                 docker_manager: DockerManager):
        self.__ts_context = ts_context
        self.__is_dev_mode = is_dev_mode
        self.__root = root
        self.__defaults_paths = defaults_paths
        self.__configs = configs
        self.__docker_manager = docker_manager

    def get_mock(self, mock_name: str) -> Optional[IMockable]:
        config = self.__configs.get(mock_name, {})
        mock_defaults_paths = self.__defaults_paths.get(mock_name, None)
        mock = None

        if mock_name == 'http':
            defaults_configurer = FileConfigurer(mock_defaults_paths)
            runnable = DockerHttp(defaults_configurer, HttpHealthCheck(), DockerHttpSetup(), DockerHttpVerify(),
                                  self.__docker_manager, DockerNetwork())
            mock = RunnableDockerMock('http', 'HTTP', runnable)
        elif mock_name == 'rabbitmq':
            defaults_configurer = FileConfigurer(mock_defaults_paths)
            configurer = BasicConfigurer(IRabbitmqBehavior.DEFAULT_CONFIG)
            configurer.merge_config(config)
            context = DockerRabbitmqContext()
            setup = DockerRabbitmqSetup(context)
            verify = DockerRabbitmqVerify(context)
            runnable = DockerRabbitmq(self.__ts_context, defaults_configurer, configurer, HttpHealthCheck(), setup,
                                      verify, self.__docker_manager, DockerNetwork())
            mock = RunnableDockerMock('rabbitmq', 'Rabbit MQ', runnable)
        elif mock_name == 'mongodb':
            defaults_configurer = FileConfigurer(mock_defaults_paths)
            context = DockerMongoContext()
            setup = DockerMongodbSetup(context)
            verify = DockerMongodbVerify(context)
            runnable = DockerMongodb(defaults_configurer, self.__is_dev_mode, setup, verify, self.__docker_manager,
                                     DockerNetwork())
            mock = RunnableDockerMock('mongodb', 'Mongo DB', runnable)
        elif mock_name == 'mysql':
            defaults_configurer = FileConfigurer(mock_defaults_paths)
            configurer = BasicConfigurer(IMysqlBehavior.DEFAULT_CONFIG)
            configurer.merge_config(config)
            context = DockerMysqlContext()
            setup = DockerMysqlSetup(context)
            verify = DockerMysqlVerify(context)
            runnable = DockerMysql(self.__ts_context, defaults_configurer, context, self.__is_dev_mode, configurer,
                                   setup, verify, self.__docker_manager, DockerNetwork())
            mock = RunnableDockerMock('mysql', 'MySQL', runnable)
        elif mock_name == 's3':
            defaults_configurer = FileConfigurer(mock_defaults_paths)
            base_objects_path = os.path.join(self.__root, 'defaults')
            setup = DockerS3Setup()
            verify = DockerS3Verify()
            runnable = DockerS3(defaults_configurer, base_objects_path, HttpHealthCheck(), setup, verify,
                                self.__docker_manager, DockerNetwork())
            mock = RunnableDockerMock('s3', 'S3', runnable)
        elif mock_name == 'filesystem':
            defaults_configurer = FileConfigurer(mock_defaults_paths)
            files_path = os.path.join(self.__root, 'io')
            base_files_path = os.path.join(self.__root, 'defaults', 'filesystem')
            setup = LocalFilesystemSetup(files_path, base_files_path)
            verify = LocalFilesystemVerify(files_path)
            runnable = LocalFilesystem(defaults_configurer, files_path, setup, verify)
            mock = RunnableLocalMock('filesystem', 'Filesystem', runnable)
        elif mock_name == 'redis':
            defaults_configurer = FileConfigurer(mock_defaults_paths)
            setup = DockerRedisSetup()
            verify = DockerRedisVerify()
            runnable = DockerRedis(defaults_configurer, self.__is_dev_mode, setup, verify, self.__docker_manager,
                                   DockerNetwork())
            mock = RunnableDockerMock('redis', 'Redis', runnable)
        return mock



from touchstone.lib import exceptions
from touchstone.lib.configurers.BasicConfigurer import BasicConfigurer
from touchstone.lib.configurers.i_configurable import IConfigurable


class TouchstoneConfig(IConfigurable):
    def __init__(self, root):
        self.__basic_configurer = BasicConfigurer({
            'root': root,
            'services': [],
            'mocks': []
        })

    def get_config(self) -> dict:
        return self.__basic_configurer.get_config()

    def merge_config(self, other: dict):
        self.__basic_configurer.merge_config(other)

    def verify(self):
        if self.__basic_configurer.get_config()['services'] is None:
            raise exceptions.TouchstoneException(f'"services" has not been set. Check your "touchstone.yml".')
        if self.__basic_configurer.get_config()['mocks'] is None:
            raise exceptions.TouchstoneException(f'"mocks" has not been set. Check your "touchstone.yml".')



import os
from typing import Optional

from touchstone.lib.configurers.service_config import ServiceConfig
from touchstone.lib.health_checks.blocking_health_check import BlockingHealthCheck
from touchstone.lib.health_checks.docker_health_check import DockerHealthCheck
from touchstone.lib.health_checks.http_health_check import HttpHealthCheck
from touchstone.lib.managers.docker_manager import DockerManager
from touchstone.lib.networking.docker_network import DockerNetwork
from touchstone.lib.nodes.services.docker.docker_executable_service import DockerExecutableService
from touchstone.lib.nodes.services.docker.docker_runnable_service import DockerRunnableService
from touchstone.lib.nodes.services.i_service import IService
from touchstone.lib.tests import Tests


class ServiceFactory(object):
    def __init__(self, is_dev_mode: bool, root: str, docker_manager: DockerManager,
                 log_directory: Optional[str]):
        self.__is_dev_mode = is_dev_mode
        self.__root = root
        self.__docker_manager = docker_manager
        self.__log_directory = log_directory

    def get_service(self, service_config: ServiceConfig, tests: Optional[Tests]) -> Optional[IService]:
        dockerfile_path = None
        if service_config.get_config()['dockerfile']:
            dockerfile_path = os.path.abspath(os.path.join(self.__root, service_config.get_config()['dockerfile']))
        service: Optional[IService] = None

        if service_config.get_config()['type'] == 'networked':
            health_check = HttpHealthCheck() if service_config.get_config()[
                                                    'availability_endpoint'] is not None else DockerHealthCheck()
            service = DockerRunnableService(service_config.get_config()['name'],
                                            tests,
                                            dockerfile_path,
                                            service_config.get_config()['docker_image'],
                                            service_config.get_config()['docker_options'],
                                            self.__docker_manager,
                                            service_config.get_config()['port'],
                                            service_config.get_config()['availability_endpoint'],
                                            self.__log_directory,
                                            DockerNetwork(),
                                            BlockingHealthCheck(service_config.get_config()['seconds_between_retries'],
                                                                service_config.get_config()['num_retries'],
                                                                health_check))
        elif service_config.get_config()['type'] == 'executable':
            service = DockerExecutableService(service_config.get_config()['name'],
                                              self.__is_dev_mode,
                                              tests,
                                              dockerfile_path,
                                              service_config.get_config()['docker_image'],
                                              service_config.get_config()['docker_options'],
                                              self.__docker_manager,
                                              service_config.get_config()['develop_command'],
                                              self.__log_directory)
        return service



from touchstone.lib import exceptions
from touchstone.lib.networking.i_network import INetwork
from touchstone.lib.nodes.mocks.behaviors.i_behavior import IBehavior
from touchstone.lib.nodes.mocks.docker.i_runnable_docker import IRunnableDocker
from touchstone.lib.nodes.mocks.i_mockable import IMockable


class RunnableDockerMock(IMockable, IRunnableDocker):
    def __init__(self, name: str, pretty_name: str, runnable_docker: IRunnableDocker):
        self.__name = name
        self.__pretty_name = pretty_name
        self.__runnable_docker = runnable_docker
        self.__has_initialized = False

    def get_name(self) -> str:
        return self.__name

    def get_pretty_name(self) -> str:
        return self.__pretty_name

    def get_behavior(self) -> IBehavior:
        return self.__runnable_docker.get_behavior()

    def initialize(self):
        self.__runnable_docker.initialize()

    def start(self):
        self.__runnable_docker.start()

    def stop(self):
        self.__runnable_docker.stop()

    def reset(self):
        self.__runnable_docker.reset()

    def get_network(self) -> INetwork:
        return self.__runnable_docker.get_network()

    def is_healthy(self) -> bool:
        try:
            self.__runnable_docker.get_network()
        except exceptions.MockException:
            return False
        is_healthy = self.__runnable_docker.is_healthy()
        if is_healthy and not self.__has_initialized:
            self.__runnable_docker.initialize()
            self.__has_initialized = True
        return is_healthy



ANY = object()


def __equals(expected, actual) -> bool:
    if expected == ANY:
        return True
    return expected == actual


def __equals_list(expected: list, actual: list) -> bool:
    if len(expected) != len(actual):
        return False
    for expected_item in expected:
        expected_item_in_actual = False
        for actual_item in actual:
            if matches(expected_item, actual_item, quiet = True):
                expected_item_in_actual = True
                break
        if not expected_item_in_actual:
            return False
    return True


def __equals_dict(expected: dict, actual: dict) -> bool:
    if len(expected) != len(actual):
        return False
    for k, v in expected.items():
        if k not in actual:
            return False
        if not matches(v, actual[k], quiet = True):
            return False
    return True


def matches(expected, actual, quiet: bool = False) -> bool:
    if isinstance(expected, dict) and isinstance(actual, dict):
        result = __equals_dict(expected, actual)
    elif isinstance(expected, list) and isinstance(actual, list):
        result = __equals_list(expected, actual)
    else:
        result = __equals(expected, actual)
    if not result and not quiet:
        print(f'Expected:\n{expected}\ndoes not match actual:\n{actual}')
    return result


def contains(expected, actual, quiet: bool = False) -> bool:
    if expected in actual:
        return True
    if not quiet:
        print(f'Expected:\n{expected}\nwas not found in actual:\n{actual}')
    return False



from typing import KeysView


class DockerMongoContext(object):
    def __init__(self):
        self.__databases = {}

    def databases(self) -> KeysView:
        return self.__databases.keys()

    def collections(self, database: str) -> list:
        return self.__databases[database]

    def add_database(self, database: str):
        self.__databases[database] = []

    def add_collection(self, database: str, collection: str):
        self.__databases[database].append(collection)

    def database_exists(self, database: str, print_warning=True) -> bool:
        if database in self.__databases:
            return True
        if print_warning:
            print(f'This database: "{database}" is not defined. Check your "mongodb.yml" defaults.')
        return False

    def collection_exists(self, database: str, collection: str, print_warning=True) -> bool:
        if database in self.__databases and collection in self.__databases[database]:
            return True
        if print_warning:
            print(f'This database: "{database}", collection: "{collection}" is not defined. Check your "mongodb.yml" '
                  f'defaults.')
        return False

    def clear(self):
        self.__databases = {}



import os
import subprocess
from typing import List, Tuple, Optional

from touchstone.lib import exceptions
from touchstone.lib.managers.docker_manager import DockerManager
from touchstone.lib.nodes.services.i_executable import IExecutable
from touchstone.lib.nodes.services.i_service import IService
from touchstone.lib.nodes.services.i_testable import ITestable
from touchstone.lib.tests import Tests


class DockerExecutableService(IService, ITestable, IExecutable):
    def __init__(self, name: str, is_dev_mode: bool, tests: Optional[Tests], dockerfile_path: Optional[str],
                 docker_image: Optional[str], docker_options: Optional[str], docker_manager: DockerManager,
                 develop_command: str, log_directory: Optional[str]):
        self.__name = name
        self.__is_dev_mode = is_dev_mode
        self.__tests = tests
        self.__dockerfile_path = dockerfile_path
        self.__docker_image = docker_image
        self.__docker_options = docker_options
        self.__docker_manager = docker_manager
        self.__develop_command = develop_command
        self.__log_directory = log_directory

    def get_name(self):
        return self.__name

    def run_test(self, file_name, test_name) -> bool:
        if not self.__tests:
            return True
        return self.__tests.run(file_name, test_name)

    def run_all_tests(self) -> bool:
        if not self.__tests:
            return True
        self.__log('Running all tests...')
        did_pass = self.__tests.run_all()
        self.__log('Finished running all tests.\n')
        return did_pass

    def execute(self, environment_vars: List[Tuple[str, str]] = []):
        if self.__is_dev_mode:
            subprocess.run(self.__develop_command, shell=True)
        else:
            if self.__dockerfile_path is None and self.__docker_image is None:
                raise exceptions.ServiceException(
                    f'{self.get_name()} could not be executed. A Dockerfile or Docker image was not supplied. Check '
                    f'your "touchstone.yml".')
            docker_artifact = None
            log_path = None
            if self.__log_directory:
                log_path = os.path.join(self.__log_directory, f'{self.__name}.log')
            if self.__dockerfile_path is not None:
                self.__log('Building and running Dockerfile...')
                docker_artifact = self.__docker_manager.build_dockerfile(self.__dockerfile_path)
            if self.__docker_image is not None:
                self.__log('Running Docker image...')
                docker_artifact = self.__docker_image
            self.__docker_manager.run_foreground_image(docker_artifact, '"$(pwd)"/touchstone/io:/app/touchstone/io',
                                                       environment_vars, log_path, self.__docker_options)

    def __log(self, message: str):
        print(f'{self.get_name()} :: {message}')



import shutil

from touchstone.lib.nodes.mocks.behaviors.i_filesystem_behavior import IFilesystemSetup


class LocalFilesystemSetup(IFilesystemSetup):
    def __init__(self, files_path: str, base_files_path: str):
        self.__files_path = files_path
        self.__base_files_path = base_files_path

    def reset(self):
        self.delete_defaults()
        shutil.copytree(self.__base_files_path, self.__files_path)

    def delete_defaults(self):
        shutil.rmtree(self.__files_path, ignore_errors=True)



import os

from touchstone import __version__


def execute():
    if os.path.exists(os.path.join('touchstone', 'touchstone.yml')):
        print('Touchstone has already been initialized.')
        exit(1)
    os.makedirs(os.path.join('touchstone', 'defaults'))
    os.makedirs(os.path.join('touchstone', 'tests'))
    open(os.path.join('touchstone', 'tests', '__init__.py'), 'a').close()
    with open(os.path.join('touchstone', 'touchstone.yml'), 'w', encoding='utf-8') as file:
        data = f"""---
touchstone_version: {__version__}
services:
  - name: my-app
    port: 8080
mocks:
"""
        file.writelines(data)
    open(os.path.join('touchstone', 'defaults', '.gitkeep'), 'a').close()
    print('Touchstone has been initialized in the current directory.')



import abc

from touchstone.lib.nodes.mocks.mocks import Mocks
from touchstone.lib.nodes.services.i_service_executor import IServiceExecutor


class TouchstoneTest(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, service_url: str, mocks: Mocks, service_executor: IServiceExecutor):
        self.service_url = service_url
        self.mocks = mocks
        self.service_executor = service_executor

    def processing_period(self) -> float:
        """The amount of time to wait before "then" is executed. This is useful to configure if you are using a mock
        or service that does not operate synchronously (e.g. Rabbit MQ)."""
        return 0.0

    @abc.abstractmethod
    def given(self) -> object:
        """GIVEN this data and/or state. Set the state of the system here."""

    @abc.abstractmethod
    def when(self, given) -> object:
        """WHEN we run this test. Your custom test code goes here. Return the result of the test to be verified."""
        return None

    @abc.abstractmethod
    def then(self, given, result) -> bool:
        """THEN check that the result is correct. Return True/False for a passed/failed test respectively."""
        return False



from abc import ABC

from touchstone.lib.health_checks.i_health_checkable import IHealthCheckable
from touchstone.lib.nodes.i_networkable import INetworkable
from touchstone.lib.nodes.mocks.behaviors.i_behavior import IBehavior
from touchstone.lib.nodes.mocks.i_runnable import IRunnable


class IRunnableDocker(ABC, IRunnable, INetworkable, IHealthCheckable):
    def get_behavior(self) -> IBehavior:
        pass



import glob
import json
import os
from typing import Optional

from touchstone.helpers import validation
from touchstone.lib.nodes.mocks.behaviors.i_filesystem_behavior import IFilesystemVerify


class LocalFilesystemVerify(IFilesystemVerify):
    def __init__(self, files_path: str):
        self.__files_path = files_path

    def file_exists(self, path: str, expected_num_files: int = 1) -> bool:
        path = os.path.join(self.__files_path, path)
        actual_num_files = len(glob.glob(path))
        if expected_num_files != actual_num_files:
            print(f'Expected number of files: "{expected_num_files}" does not match actual: "{actual_num_files}" '
                  f'at path: "{path}".')
            return False
        return True

    def file_matches(self, path: str, expected: bytes) -> bool:
        path = self.get_file_matching(path)
        if not path:
            print(f'Expected exactly 1 matching file at: "{path}".')
            return False
        with open(path, 'rb') as data:
            data_matches = bytes(data.read()) == expected
        if not data_matches:
            print('Expected data does not match data in file.')
        return data_matches

    def file_matches_json(self, path: str, expected: dict) -> bool:
        path = self.get_file_matching(path)
        if not path:
            print(f'Expected exactly 1 matching file at: "{path}".')
            return False
        with open(path, 'rb') as data:
            actual = json.loads(data.read().decode('utf-8'))
        return validation.matches(expected, actual)

    def get_file_matching(self, path: str) -> Optional[str]:
        path = os.path.join(self.__files_path, path)
        all_files = glob.glob(path)
        if len(all_files) != 1:
            return None
        return all_files[0]



import glob
import os
import sys
from pathlib import Path

import yaml

from touchstone.lib import exceptions
from touchstone.lib.configurers.service_config import ServiceConfig
from touchstone.lib.configurers.touchstone_config import TouchstoneConfig
from touchstone.lib.managers.docker_manager import DockerManager
from touchstone.lib.nodes.mocks.mock_factory import MockFactory
from touchstone.lib.nodes.mocks.mocks import Mocks
from touchstone.lib.nodes.services.service_factory import ServiceFactory
from touchstone.lib.nodes.services.services import Services
from touchstone.lib.tests import Tests
from touchstone.lib.ts_context import TsContext


class Bootstrap(object):
    def __init__(self, is_dev_mode=False, should_log_services=False):
        root = os.path.join(os.getcwd(), 'touchstone')
        touchstone_config = self.__build_touchstone_config(root)
        log_directory = None
        if should_log_services:
            log_directory = os.path.join(root, 'logs')
            os.makedirs(log_directory, exist_ok=True)
        self.ts_context = TsContext()
        self.docker_manager = DockerManager(should_auto_discover=not is_dev_mode)
        self.mocks = self.__build_mocks(is_dev_mode,
                                        touchstone_config.get_config()['root'],
                                        touchstone_config.get_config()['mocks'])
        self.services = self.__build_services(is_dev_mode,
                                              touchstone_config.get_config()['root'],
                                              touchstone_config.get_config()['services'],
                                              self.mocks, log_directory)

    def __build_touchstone_config(self, root) -> TouchstoneConfig:
        config = TouchstoneConfig(root)
        path = os.path.join(root, 'touchstone.yml')
        with open(path, 'r') as file:
            config.merge_config(yaml.safe_load(file))
        config.verify()
        return config

    def __build_mocks(self, is_dev_mode, root, configs) -> Mocks:
        defaults_paths = {}
        default_files = glob.glob(os.path.join(root, 'defaults') + '/*.yml')
        for default_file in default_files:
            defaults_paths[Path(default_file).stem] = default_file

        mocks = Mocks()
        mock_factory = MockFactory(self.ts_context, is_dev_mode, root, defaults_paths, configs, self.docker_manager)
        for mock_name in configs:
            mock = mock_factory.get_mock(mock_name)
            if not mock:
                raise exceptions.MockNotSupportedException(f'Mock: {mock_name} is not supported.')
            setattr(mocks, mock_name, mock.get_behavior())
            mocks.register_mock(mock)
        return mocks

    def __build_services(self, is_dev_mode, root, user_service_configs, mocks, log_directory) -> Services:
        services = Services()
        service_factory = ServiceFactory(is_dev_mode, root, self.docker_manager, log_directory)
        for user_service_config in user_service_configs:
            service_config = ServiceConfig()
            service_config.merge_config(user_service_config)
            service_config.verify()
            if service_config.get_config()['tests'] is None:
                tests = None
            else:
                tests_path = os.path.abspath(os.path.join(root, service_config.get_config()['tests']))
                tests = Tests(mocks, services, tests_path)
            service = service_factory.get_service(service_config, tests)
            services.add_service(service)
        return services

    def cleanup(self):
        self.services.stop()
        self.mocks.stop()
        self.docker_manager.cleanup()

    def exit(self, is_successful: bool):
        print('Shutting down...')
        if is_successful:
            code = 0
        else:
            code = 1
        self.cleanup()
        sys.exit(code)



import glob
import importlib.util
import inspect
import os
import sys

from touchstone.lib.nodes.mocks.mocks import Mocks
from touchstone.lib.nodes.services.i_service_executor import IServiceExecutor
from touchstone.lib.test import TestContainer, TestClass


class Tests(object):
    __TEST_PREFIX = 'test_'

    def __init__(self, mocks: Mocks, service_executor: IServiceExecutor, tests_path: str):
        self.__mocks = mocks
        self.__service_executor = service_executor
        self.__tests_path = tests_path

    def run(self, file_name: str, test_name: str, service_url: str = None) -> bool:
        self.__load_package(self.__tests_path)
        self.__reload_support_modules(self.__tests_path)
        test_container = self.__load_test_class(file_name, test_name)

        if not test_container:
            print(f'No tests found with file name "{file_name}".')
            return False
        return test_container.execute(service_url, self.__mocks, self.__service_executor)

    def run_all(self, service_url: str = None) -> bool:
        self.__load_package(self.__tests_path)
        self.__reload_support_modules(self.__tests_path)
        all_test_containers = self.__load_test_classes()
        if len(all_test_containers) == 0:
            print(f'No tests found at {self.__tests_path}')
            return False

        tests_passed = True
        for test_container in all_test_containers:
            print(f'\n{test_container.file}')
            if not test_container.execute(service_url, self.__mocks, self.__service_executor):
                tests_passed = False

        self.__mocks.reset()
        return tests_passed

    def __load_module(self, name: str, path):
        spec = importlib.util.spec_from_file_location(name, path)
        module = importlib.util.module_from_spec(spec)
        sys.modules[spec.name] = module
        spec.loader.exec_module(module)
        return module

    def __load_package(self, path: str):
        module_name = os.path.basename(path)
        init_path = os.path.join(path, '__init__.py')
        self.__load_module(module_name, init_path)

    def __reload_support_modules(self, path: str):
        package_len = len(path) - len(os.path.basename(path))
        files = glob.glob(path + '/**/*.py', recursive=True)
        for file in files:
            file_name = os.path.basename(file)
            if self.__TEST_PREFIX not in file_name and '__init__' not in file_name:
                module = file[package_len:].replace('/', '.')[:-3]
                if module in sys.modules:
                    importlib.reload(sys.modules[module])

    def __load_test_class(self, file_name, test_name):
        file = glob.glob(os.path.join(self.__tests_path, file_name + '.py'))
        if len(file) == 0:
            return None
        file = file[0]
        test_container = TestContainer(file)
        module = self.__load_module(file, file)
        members = [o for o in inspect.getmembers(module) if inspect.isclass(o[1])]
        for member in members:
            class_name = member[0]
            class_type = member[1]
            if class_name.lower() == test_name and module.__file__ == class_type.__module__:
                test_class = TestClass(class_name, class_type)
                test_container.add_test_class(test_class)
        return test_container

    def __load_test_classes(self):
        files = glob.glob(self.__tests_path + f'/**/{self.__TEST_PREFIX}*.py', recursive=True)
        all_test_containers = []
        for file in files:
            test_container = TestContainer(file)
            all_test_containers.append(test_container)
            module = self.__load_module(file, file)
            members = [o for o in inspect.getmembers(module) if inspect.isclass(o[1])]
            for member in members:
                class_name = member[0]
                class_type = member[1]
                if module.__file__ == class_type.__module__:
                    test_class = TestClass(class_name, class_type)
                    test_container.add_test_class(test_class)
        return all_test_containers



import json

from redis import Redis

from touchstone.helpers import validation
from touchstone.lib.nodes.mocks.behaviors.i_redis_behavior import IRedisVerify


class DockerRedisVerify(IRedisVerify):
    def __init__(self):
        self.__redis_client: Redis = None

    def set_redis_client(self, redis_client: Redis):
        self.__redis_client = redis_client

    def value_exists(self, key: str) -> bool:
        return self.__redis_client.get(key) is not None

    def value_matches(self, key: str, value: str) -> bool:
        return self.__redis_client.get(key).decode('utf-8') == value

    def value_matches_json(self, key: str, value: dict) -> bool:
        value_json = json.loads(self.__redis_client.get(key).decode('utf-8'))
        return validation.matches(value, value_json)



import json

from minio import Minio

from touchstone.helpers import validation
from touchstone.lib.nodes.mocks.behaviors.i_s3_behavior import IS3Verify


class DockerS3Verify(IS3Verify):
    def __init__(self):
        self.__s3_client = None

    def set_s3_client(self, s3_client: Minio):
        self.__s3_client = s3_client

    def bucket_exists(self, name: str) -> bool:
        return self.__s3_client.bucket_exists(name)

    def object_exists(self, bucket_name: str, object_name: str) -> bool:
        exists = self.__s3_client.stat_object(bucket_name, object_name)
        if not exists:
            print(f'Object name: "{object_name}" could not be found in bucket: "{bucket_name}"')
        return exists

    def object_matches(self, bucket_name: str, object_name: str, expected: bytes) -> bool:
        data = self.__s3_client.get_object(bucket_name, object_name).data
        data_matches = data == expected
        if not data_matches:
            print('Expected data does not match data in object.')
        return data_matches

    def object_matches_json(self, bucket_name: str, object_name: str, expected_json: dict) -> bool:
        data = self.__s3_client.get_object(bucket_name, object_name).data
        actual = json.loads(data.decode('utf-8'))
        return validation.matches(expected_json, actual)



from touchstone.lib import exceptions
from touchstone.lib.configurers.BasicConfigurer import BasicConfigurer
from touchstone.lib.configurers.i_configurable import IConfigurable


class ServiceConfig(IConfigurable):
    def __init__(self):
        self.__basic_configurer = BasicConfigurer({
            'type': 'networked',
            'name': 'unnamed-service',
            'tests': './tests',
            'port': 8080,
            'dockerfile': None,
            'docker_image': None,
            'docker_options': None,
            'availability_endpoint': None,
            'num_retries': 20,
            'seconds_between_retries': 5,
            'develop_command': ''
        })

    def get_config(self) -> dict:
        return self.__basic_configurer.get_config()

    def merge_config(self, other: dict):
        self.__basic_configurer.merge_config(other)

    def verify(self):
        service_type = self.__basic_configurer.get_config()['type']
        if service_type != 'networked' and service_type != 'executable':
            raise exceptions.TouchstoneException(
                f'Unknown service type: "{service_type}". Check your "touchstone.yml".')
        if self.__basic_configurer.get_config()['dockerfile'] and self.__basic_configurer.get_config()['docker_image']:
            raise exceptions.TouchstoneException(
                '"dockerfile" and "docker_image" cannot both be set. Check your "touchstone.yml".')



from redis import Redis

from touchstone.lib import exceptions
from touchstone.lib.nodes.mocks.behaviors.i_redis_behavior import IRedisSetup


class DockerRedisSetup(IRedisSetup):
    def __init__(self):
        self.__redis_client: Redis = None

    def set_redis_client(self, redis_client: Redis):
        self.__redis_client = redis_client

    def init(self, defaults: dict):
        self.__redis_client.flushall()
        for key, value in defaults.get('objects', []).items():
            self.__redis_client.set(key, value)

    def set(self, key: str, value: str):
        if not isinstance(value, str):
            raise exceptions.MockException('Value must be a str.')
        return self.__redis_client.set(key, value)



from touchstone.lib.networking.i_network import INetwork


class DockerNetwork(INetwork):
    def __init__(self):
        self.__external_host = 'localhost'
        self.__container_id = None
        self.__internal_port = None
        self.__external_port = None
        self.__ui_port = None
        self.__ui_endpoint = ''
        self.__prefix = ''
        self.__endpoint = ''
        self.__username = None
        self.__password = None

    def internal_host(self) -> str:
        return self.__container_id

    def external_host(self) -> str:
        return self.__external_host

    def container_id(self) -> str:
        return self.__container_id

    def set_container_id(self, container_id: str):
        self.__container_id = container_id

    def internal_port(self) -> int:
        return self.__internal_port

    def set_internal_port(self, internal_port: int):
        self.__internal_port = internal_port

    def external_port(self) -> int:
        return self.__external_port

    def set_external_port(self, external_port: int):
        self.__external_port = external_port

    def set_ui_port(self, ui_port: int):
        self.__ui_port = ui_port

    def set_ui_endpoint(self, ui_endpoint: str):
        self.__ui_endpoint = ui_endpoint

    def set_prefix(self, prefix: str):
        self.__prefix = prefix

    def set_endpoint(self, endpoint: str):
        self.__endpoint = endpoint

    def username(self) -> str:
        return self.__username

    def set_username(self, username: str):
        self.__username = username

    def password(self) -> str:
        return self.__password

    def set_password(self, password: str):
        self.__password = password

    def internal_url(self) -> str:
        return f'{self.__prefix + self.__container_id}:{self.__internal_port}{self.__endpoint}'

    def external_url(self) -> str:
        return f'{self.__prefix + self.__external_host}:{self.__external_port}{self.__endpoint}'

    def ui_url(self) -> str:
        return f'http://{self.__external_host}:{self.__ui_port}{self.__ui_endpoint}'



import json
import urllib.request
from json.decoder import JSONDecodeError
from typing import Optional


def __request(url: str, method: str, body: Optional[str], headers: Optional[dict]) -> str:
    headers = headers or {}
    body = bytes(body, encoding='utf-8') if body else None
    request = urllib.request.Request(url, method=method, data=body, headers=headers)
    return urllib.request.urlopen(request).read().decode('utf-8')


def __decode_json(json_str: str) -> Optional[dict]:
    try:
        return json.loads(json_str)
    except JSONDecodeError:
        print(f'Response JSON: "{json_str}" is not valid.')
        return None


def get(url: str, headers: dict = None) -> str:
    """Makes a GET request to the specified URL."""
    return __request(url, 'GET', None, headers)


def get_json(url: str, headers: dict = None) -> dict:
    """Makes a GET request to the specified URL returning JSON."""
    return __decode_json(get(url, headers))


def post(url: str, body: str = None, headers: dict = None) -> str:
    """Makes a POST request to the specified URL with a body."""
    return __request(url, 'POST', body, headers)


def post_json(url: str, body: dict = None, headers: dict = None) -> dict:
    """Makes a POST request to the specified URL with a JSON body returning JSON."""
    headers = headers or {}
    headers['Content-type'] = 'application/json'
    return __decode_json(post(url, json.dumps(body), headers))


def put(url: str, body: str = None, headers: dict = None) -> str:
    """Makes a PUT request to the specified URL with a body."""
    return __request(url, 'PUT', body, headers)


def put_json(url: str, body: dict = None, headers: dict = None) -> dict:
    """Makes a PUT request to the specified URL with a JSON body returning JSON."""
    headers = headers or {}
    headers['Content-type'] = 'application/json'
    return __decode_json(put(url, json.dumps(body), headers))


def delete(url: str, body: str = None, headers: dict = None) -> str:
    """Makes a DELETE request to the specified URL with a body."""
    return __request(url, 'DELETE', body, headers)


def delete_json(url: str, body: dict = None, headers: dict = None) -> dict:
    """Makes a DELETE request to the specified URL with a JSON body returning JSON."""
    headers = headers or {}
    headers['Content-type'] = 'application/json'
    return __decode_json(delete(url, json.dumps(body), headers))



from pyfiglet import figlet_format

from touchstone import common
from touchstone.bootstrap import Bootstrap


def execute(should_log_services=False):
    if not common.sanity_check_passes():
        exit(1)
    bootstrap = Bootstrap(False, should_log_services)
    print(figlet_format('Touchstone', font='larry3d'))

    try:
        bootstrap.mocks.start()
        bootstrap.services.add_environment_vars(bootstrap.mocks.environment_vars())
        bootstrap.services.start()
        bootstrap.ts_context.trigger_services_available()
        tests_did_pass = bootstrap.services.run_all_tests()
        if tests_did_pass:
            print('All Touchstone tests passed successfully!')
        else:
            print('One or more Touchstone tests failed.')
        bootstrap.exit(tests_did_pass)
    except (Exception, KeyboardInterrupt) as e:
        print('\nTouchstone was interrupted. Cleaning up...')
        bootstrap.cleanup()
        raise e



from touchstone.lib.nodes.mocks.behaviors.i_behavior import IBehavior
from touchstone.lib.nodes.mocks.i_mockable import IMockable
from touchstone.lib.nodes.mocks.local.i_runnable_local import IRunnableLocal


class RunnableLocalMock(IMockable, IRunnableLocal):
    def __init__(self, name: str, pretty_name: str, runnable_local: IRunnableLocal):
        self.__name = name
        self.__pretty_name = pretty_name
        self.__runnable_local = runnable_local

    def get_name(self) -> str:
        return self.__name

    def get_pretty_name(self) -> str:
        return self.__pretty_name

    def get_behavior(self) -> IBehavior:
        return self.__runnable_local.get_behavior()

    def initialize(self):
        self.__runnable_local.initialize()

    def start(self):
        self.__runnable_local.start()

    def stop(self):
        self.__runnable_local.stop()

    def reset(self):
        self.__runnable_local.reset()

    def is_healthy(self) -> bool:
        return self.__runnable_local.is_healthy()



import abc

from touchstone.lib.nodes.mocks.behaviors.i_behavior import IBehavior


class IRedisSetup(object):
    @abc.abstractmethod
    def set(self, key: str, value: str):
        """Sets a key-value pair."""
        pass


class IRedisVerify(object):
    @abc.abstractmethod
    def value_exists(self, key: str) -> bool:
        """Returns True if the given key exists."""
        pass

    @abc.abstractmethod
    def value_matches(self, key: str, value: str) -> bool:
        """Returns True if the given key matches the given value."""

    @abc.abstractmethod
    def value_matches_json(self, key: str, value: dict) -> bool:
        """Returns True if the given key matches the given JSON value."""


class IRedisBehavior(IBehavior):
    @abc.abstractmethod
    def setup(self) -> IRedisSetup:
        pass

    @abc.abstractmethod
    def verify(self) -> IRedisVerify:
        pass



import abc


class INetwork(object):
    @abc.abstractmethod
    def internal_host(self) -> str:
        pass

    @abc.abstractmethod
    def external_host(self) -> str:
        pass

    @abc.abstractmethod
    def internal_port(self) -> int:
        pass

    @abc.abstractmethod
    def external_port(self) -> int:
        pass

    @abc.abstractmethod
    def username(self) -> str:
        pass

    @abc.abstractmethod
    def password(self) -> str:
        pass

    @abc.abstractmethod
    def internal_url(self) -> str:
        pass

    @abc.abstractmethod
    def external_url(self) -> str:
        pass

    @abc.abstractmethod
    def ui_url(self) -> str:
        pass



from touchstone import common


class DockerRabbitmqContext(object):
    def __init__(self):
        self.__exchanges: dict = {}

    def track(self, exchange: str, routing_key: str):
        if exchange not in self.__exchanges:
            self.__exchanges[exchange] = {}
        self.__exchanges[exchange][routing_key] = {
            'times': 0,
            'payloads': []
        }
        common.logger.debug(f'Now tracking: {self.__exchanges}')

    def track_payload(self, exchange: str, routing_key: str, payload: str):
        self.__exchanges[exchange][routing_key]['times'] += 1
        self.__exchanges[exchange][routing_key]['payloads'].append(payload)

    def exchange_is_tracked(self, exchange: str, routing_key: str, print_warning=True) -> bool:
        if exchange in self.__exchanges and routing_key in self.__exchanges[exchange]:
            return True
        if print_warning:
            print(f'This exchange: "{exchange}", routing-key: "{routing_key}" combination is not defined. Check your '
                  f'"rabbitmq.yml" defaults.')
        return False

    def messages_published(self, exchange: str, routing_key: str) -> int:
        return self.__exchanges[exchange][routing_key]['times']

    def payloads_published(self, exchange: str, routing_key: str) -> list:
        return self.__exchanges[exchange][routing_key]['payloads']

    def clear(self):
        for exchange in self.__exchanges:
            for routing_key in self.__exchanges[exchange]:
                self.__exchanges[exchange][routing_key]['times'] = 0
                self.__exchanges[exchange][routing_key]['payloads'] = []



import subprocess
import urllib.error
import urllib.request

from touchstone import common
from touchstone.lib import exceptions
from touchstone.lib.health_checks.i_health_checkable import IHealthCheckable


class DockerHealthCheck(IHealthCheckable):
    def __init__(self):
        self.__container_id = None

    def set_container_id(self, container_id: str):
        self.__container_id = container_id

    def is_healthy(self) -> bool:
        if not self.__container_id:
            raise exceptions.MockException('Container ID must be set before checking health.')
        try:
            common.logger.debug(f'Executing Docker health check for container ID: {self.__container_id}')
            passed = 'healthy' in str(
                subprocess.run(['docker', 'inspect', "--format='{{.State.Health.Status}}'", self.__container_id],
                               stdout=subprocess.PIPE).stdout, encoding='utf-8')
            if passed:
                common.logger.debug(f'Docker health check passed for container ID: {self.__container_id}')
            else:
                common.logger.debug(f'Docker health check failed for container ID: {self.__container_id}')
            return passed
        except (urllib.error.URLError, ConnectionResetError):
            return False



import json
import urllib.request
from json.decoder import JSONDecodeError

from touchstone.helpers import validation
from touchstone.lib.nodes.mocks.behaviors.i_http_behavior import IHttpVerify


class DockerHttpVerify(IHttpVerify):
    def __init__(self):
        super().__init__()
        self.__url = None

    def set_url(self, url: str):
        self.__url = url

    def get_called(self, endpoint: str, times: int = 1, url_pattern: bool = False) -> bool:
        return self.__count_verification(endpoint, 'GET', times, url_pattern)

    def post_called(self, endpoint: str, times: int = 1, url_pattern: bool = False) -> bool:
        return self.__count_verification(endpoint, 'POST', times, url_pattern)

    def post_contained(self, endpoint: str, expected_body: str, url_pattern: bool = False) -> bool:
        return self.__contained_verification(endpoint, 'POST', expected_body, url_pattern)

    def post_contained_json(self, endpoint: str, expected_body: dict, url_pattern: bool = False) -> bool:
        return self.__contained_json_verification(endpoint, 'POST', expected_body, url_pattern)

    def put_called(self, endpoint: str, times: int = 1, url_pattern: bool = False) -> bool:
        return self.__count_verification(endpoint, 'PUT', times, url_pattern)

    def put_contained(self, endpoint: str, expected_body: str, url_pattern: bool = False) -> bool:
        return self.__contained_verification(endpoint, 'PUT', expected_body, url_pattern)

    def put_contained_json(self, endpoint: str, expected_body: dict, url_pattern: bool = False) -> bool:
        return self.__contained_json_verification(endpoint, 'PUT', expected_body, url_pattern)

    def delete_called(self, endpoint: str, times: int = 1, url_pattern: bool = False) -> bool:
        return self.__count_verification(endpoint, 'DELETE', times, url_pattern)

    def delete_contained(self, endpoint: str, expected_body: str, url_pattern: bool = False) -> bool:
        return self.__contained_verification(endpoint, 'DELETE', expected_body, url_pattern)

    def delete_contained_json(self, endpoint: str, expected_body: dict, url_pattern: bool = False) -> bool:
        return self.__contained_json_verification(endpoint, 'DELETE', expected_body, url_pattern)

    def __count_verification(self, endpoint, http_verb, times, url_pattern):
        payload = {
            'method': http_verb,
            self.__get_url_type(url_pattern): endpoint
        }
        data = json.dumps(payload).encode('utf-8')
        request = urllib.request.Request(
            f'{self.__url}/__admin/requests/count',
            data=data,
            headers={'Content-Type': 'application/json'})
        response = urllib.request.urlopen(request).read()
        call_count = json.loads(response)['count']
        if times is None:
            return call_count > 0
        return validation.matches(times, call_count)

    def __get_request_bodies(self, endpoint, http_verb, url_pattern):
        payload = {
            'method': http_verb,
            self.__get_url_type(url_pattern): endpoint
        }
        data = json.dumps(payload).encode('utf-8')
        request = urllib.request.Request(
            f'{self.__url}/__admin/requests/find',
            data=data,
            headers={'Content-Type': 'application/json'})
        response = urllib.request.urlopen(request).read()
        bodies = []
        for request in json.loads(response)['requests']:
            bodies.append(request['body'])
        return bodies

    def __get_url_type(self, url_pattern: bool):
        if url_pattern:
            return 'urlPattern'
        return 'url'

    def __contained_verification(self, endpoint, http_verb, expected_body, url_pattern):
        bodies = self.__get_request_bodies(endpoint, http_verb, url_pattern)
        return validation.contains(expected_body, bodies)

    def __contained_json_verification(self, endpoint, http_verb, expected_body, url_pattern):
        bodies = self.__get_request_bodies(endpoint, http_verb, url_pattern)
        for body in bodies:
            try:
                body = json.loads(body)
                if validation.matches(expected_body, body, quiet=True):
                    return True
            except JSONDecodeError:
                pass
        print(f'Expected:\n{expected_body}\nwas not found in actual:\n{bodies}')
        return False



from typing import List, Iterable

from pymysql.cursors import Cursor

from touchstone import common
from touchstone.lib.nodes.mocks.behaviors.i_mysql_behabior import IMysqlSetup
from touchstone.lib.nodes.mocks.docker.mysql.docker_mysql_context import DockerMysqlContext


class DockerMysqlSetup(IMysqlSetup):
    def __init__(self, mysql_context: DockerMysqlContext):
        self.__mysql_context = mysql_context
        self.__cursor = None
        self.__convert_camel_to_snake = False

    def set_cursor(self, cursor: Cursor):
        self.__cursor = cursor

    def set_convert_camel_to_snake(self, convert_camel_to_snake: bool):
        self.__convert_camel_to_snake = convert_camel_to_snake

    def recreate_databases(self):
        for database in self.__mysql_context.databases:
            self.__cursor.execute(f'DROP DATABASE {database}')
            self.__cursor.execute(f'CREATE DATABASE {database} CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci')

    def init(self, defaults: dict):
        for database in self.__mysql_context.databases:
            self.__cursor.execute(f'DROP DATABASE {database}')
        self.__mysql_context.clear()

        for database in defaults.get('databases', []):
            database_name = database['name']
            self.__cursor.execute(f'CREATE DATABASE {database_name} CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci')
            self.__cursor.execute(f'USE {database_name}')
            self.__mysql_context.add_database(database_name)
            for statement in database.get('statements', []):
                self.__cursor.execute(statement)

    def execute(self, database: str, sql: str):
        if self.__mysql_context.database_exists(database):
            self.__cursor.execute(f'USE {database}')
            self.__cursor.execute(sql)

    def insert_row(self, database: str, table: str, data: dict):
        if self.__mysql_context.database_exists(database):
            if self.__convert_camel_to_snake:
                data = common.to_snake(data)
            values = self.__sql_values_from_dict(data)
            sql = self.__build_insert_sql(table, data.keys(), values)
            common.logger.debug(f'Executing: {sql}')
            self.__cursor.execute(f'USE {database}')
            self.__cursor.execute(sql, data)

    def insert_rows(self, database: str, table: str, data: List[dict]):
        for datum in data:
            self.insert_row(database, table, datum)

    def __sql_values_from_dict(self, data: dict) -> str:
        col_values = []
        for key, value in data.items():
            if value is None:
                col_values.append('NULL')
            else:
                col_values.append(f'%({key})s')
        col_values = ', '.join(col_values)
        return '(' + col_values + ')'

    def __sql_values_from_list(self, data: List[dict]) -> str:
        values = []
        for datum in data:
            values.append(self.__sql_values_from_dict(datum))
        return ', '.join(values)

    def __build_insert_sql(self, table: str, cols: Iterable, values: str) -> str:
        col_names = ', '.join(cols)
        return f"INSERT INTO {table} ({col_names}) VALUES {values}"



import pika

from touchstone.lib import exceptions
from touchstone.lib.configurers.i_configurable import IConfigurable
from touchstone.lib.health_checks.http_health_check import HttpHealthCheck
from touchstone.lib.listeners.i_services_available_listener import IServicesAvailableListener
from touchstone.lib.managers.docker_manager import DockerManager
from touchstone.lib.networking.docker_network import DockerNetwork
from touchstone.lib.networking.i_network import INetwork
from touchstone.lib.nodes.mocks.behaviors.i_behavior import IBehavior
from touchstone.lib.nodes.mocks.behaviors.i_rabbitmq_behavior import IRabbitmqBehavior, IRabbitmqVerify, \
    IRabbitmqSetup
from touchstone.lib.nodes.mocks.docker.i_runnable_docker import IRunnableDocker
from touchstone.lib.nodes.mocks.docker.rabbitmq.docker_rabbitmq_setup import DockerRabbitmqSetup
from touchstone.lib.nodes.mocks.docker.rabbitmq.docker_rabbitmq_verify import DockerRabbitmqVerify
from touchstone.lib.ts_context import TsContext


class DockerRabbitmq(IRunnableDocker, IRabbitmqBehavior, IServicesAvailableListener):
    __USERNAME = 'guest'
    __PASSWORD = 'guest'

    def __init__(self, ts_context: TsContext, defaults_configurer: IConfigurable, configurer: IConfigurable,
                 health_check: HttpHealthCheck, setup: DockerRabbitmqSetup, verify: DockerRabbitmqVerify,
                 docker_manager: DockerManager, docker_network: DockerNetwork):
        ts_context.register_services_available_listener(self)
        self.__defaults_configurer = defaults_configurer
        self.__configurer = configurer
        self.__health_check = health_check
        self.__setup = setup
        self.__verify = verify
        self.__docker_manager = docker_manager
        self.__docker_network = docker_network

    def get_behavior(self) -> IBehavior:
        return self

    def get_network(self) -> INetwork:
        return self.__docker_network

    def initialize(self):
        connection_params = pika.ConnectionParameters(
            host=self.__docker_network.external_host(),
            port=self.__docker_network.external_port(),
            credentials=pika.PlainCredentials(self.__USERNAME, self.__PASSWORD),
            heartbeat=0
        )
        connection = pika.BlockingConnection(connection_params)
        channel = connection.channel()
        self.__setup.set_channel(channel)
        self.__setup.set_connection_params(connection_params)
        self.__verify.set_blocking_channel(channel)
        if self.__configurer.get_config()['auto_create']:
            self.__setup.create_all(self.__defaults_configurer.get_config())

    def start(self):
        run_result = self.__docker_manager.run_background_image('rabbitmq:3.7.22-management-alpine', port=5672,
                                                                ui_port=15672)
        self.__docker_network.set_container_id(run_result.container_id)
        self.__docker_network.set_internal_port(run_result.internal_port)
        self.__docker_network.set_external_port(run_result.external_port)
        self.__docker_network.set_ui_port(run_result.ui_port)
        self.__docker_network.set_username(self.__USERNAME)
        self.__docker_network.set_password(self.__PASSWORD)

    def stop(self):
        if self.__docker_network.container_id():
            self.__setup.stop_listening()
            self.__docker_manager.stop_container(self.__docker_network.container_id())

    def reset(self):
        self.__setup.purge_queues()

    def is_healthy(self) -> bool:
        self.__health_check.set_url(self.__docker_network.ui_url())
        return self.__health_check.is_healthy()

    def setup(self) -> IRabbitmqSetup:
        if not self.__setup:
            raise exceptions.MockException('Setup unavailable. Mock is still starting.')
        return self.__setup

    def verify(self) -> IRabbitmqVerify:
        if not self.__verify:
            raise exceptions.MockException('Verify unavailable. Mock is still starting.')
        return self.__verify

    def services_available(self):
        if not self.__configurer.get_config()['auto_create']:
            self.__setup.create_shadow_queues(self.__defaults_configurer.get_config())



class DockerMysqlContext(object):
    def __init__(self):
        self.__databases = []

    @property
    def databases(self) -> list:
        return self.__databases.copy()

    def add_database(self, database: str):
        self.__databases.append(database)

    def database_exists(self, database: str, print_warning=True) -> bool:
        if database in self.__databases:
            return True
        if print_warning:
            print(f'This database: "{database}" is not defined. Check your "mysql.yml" defaults.')
        return False

    def clear(self):
        self.__databases = []



import os
from typing import List, Tuple, Optional

from touchstone.lib import exceptions
from touchstone.lib.health_checks.blocking_health_check import BlockingHealthCheck
from touchstone.lib.health_checks.docker_health_check import DockerHealthCheck
from touchstone.lib.health_checks.http_health_check import HttpHealthCheck
from touchstone.lib.managers.docker_manager import DockerManager
from touchstone.lib.networking.docker_network import DockerNetwork
from touchstone.lib.networking.i_network import INetwork
from touchstone.lib.nodes.i_networkable import INetworkable
from touchstone.lib.nodes.services.i_runnable import IRunnable
from touchstone.lib.nodes.services.i_service import IService
from touchstone.lib.nodes.services.i_testable import ITestable
from touchstone.lib.tests import Tests


class DockerRunnableService(IService, ITestable, IRunnable, INetworkable):
    def __init__(self, name: str, tests: Optional[Tests], dockerfile_path: Optional[str], docker_image: Optional[str],
                 docker_options: Optional[str], docker_manager: DockerManager, port: int, availability_endpoint: str,
                 log_directory: Optional[str], docker_network: DockerNetwork,
                 blocking_health_check: BlockingHealthCheck):
        self.__name = name
        self.__tests = tests
        self.__dockerfile_path = dockerfile_path
        self.__docker_image = docker_image
        self.__docker_options = docker_options
        self.__docker_manager = docker_manager
        self.__port = port
        self.__availability_endpoint = availability_endpoint
        self.__log_directory = log_directory
        self.__docker_network = docker_network
        self.__blocking_health_check = blocking_health_check

    def get_name(self):
        return self.__name

    def run_test(self, file_name, test_name) -> bool:
        if not self.__tests:
            return True
        return self.__tests.run(file_name, test_name, self.url())

    def run_all_tests(self) -> bool:
        if not self.__tests:
            return True
        self.__log('Running all tests...')
        did_pass = self.__tests.run_all(self.url())
        self.__log('Finished running all tests.\n')
        return did_pass

    def start(self, environment_vars: List[Tuple[str, str]] = []):
        if self.__dockerfile_path is None and self.__docker_image is None:
            raise exceptions.ServiceException(
                f'{self.get_name()} could not be started. A Dockerfile or Docker image was not supplied. Check your '
                f'"touchstone.yml".')
        docker_artifact = None
        if self.__dockerfile_path is not None:
            self.__log('Building and running Dockerfile...')
            docker_artifact = self.__docker_manager.build_dockerfile(self.__dockerfile_path)
        elif self.__docker_image is not None:
            self.__log('Running Docker image...')
            docker_artifact = self.__docker_image
        log_path = None
        if self.__log_directory:
            log_path = os.path.join(self.__log_directory, f'{self.__name}.log')
        run_result = self.__docker_manager.run_background_image(docker_artifact, self.__port,
                                                                environment_vars=environment_vars,
                                                                hostname=self.__name,
                                                                log_path=log_path,
                                                                options=self.__docker_options)
        self.__docker_network.set_container_id(run_result.container_id)
        self.__docker_network.set_external_port(run_result.external_port)

    def stop(self):
        if self.__docker_network.container_id():
            self.__docker_manager.stop_container(self.__docker_network.container_id())
            self.__docker_network.set_container_id(None)

    def is_running(self) -> bool:
        return self.__docker_network.container_id() is not None

    def wait_until_healthy(self):
        target = self.__blocking_health_check.get_target()
        if isinstance(target, HttpHealthCheck):
            target.set_url(self.url() + self.__availability_endpoint)
        elif isinstance(target, DockerHealthCheck):
            target.set_container_id(self.__docker_network.container_id())
        else:
            raise exceptions.ServiceException('Health check type not supported.')
        is_healthy = self.__blocking_health_check.wait_until_healthy()
        if not is_healthy:
            raise exceptions.ServiceException('Could not connect to service\'s availability endpoint.')

    def url(self):
        port = self.__docker_network.external_port() if self.is_running() else self.__port
        return f'http://{self.__docker_network.external_host()}:{port}'

    def get_network(self) -> INetwork:
        return self.__docker_network

    def __log(self, message: str):
        print(f'{self.get_name()} :: {message}')



import abc


class IRunnable(object):
    @abc.abstractmethod
    def initialize(self):
        pass

    @abc.abstractmethod
    def start(self):
        pass

    @abc.abstractmethod
    def stop(self):
        pass

    @abc.abstractmethod
    def reset(self):
        pass



import abc
from typing import Optional

from touchstone.lib.nodes.mocks.behaviors.i_behavior import IBehavior


class IFilesystemSetup(object):
    pass


class IFilesystemVerify(object):
    @abc.abstractmethod
    def file_exists(self, path: str, expected_num_files: int = 1) -> bool:
        """Returns True if the given path to a file exists. Assumes "touchstone/io" as a base directory. Wildcards may
        be used in path."""
        pass

    @abc.abstractmethod
    def file_matches(self, path: str, expected: bytes) -> bool:
        """Returns True if the given path to a file matches the expected bytes. Assumes "touchstone/io" as a base
        directory. Wildcards may be used in path."""
        pass

    @abc.abstractmethod
    def file_matches_json(self, path: str, expected: dict) -> bool:
        """Returns True if the given path to a file matches the expected JSON. Assumes "touchstone/io" as a base
        directory. Wildcards may be used in path."""
        pass

    @abc.abstractmethod
    def get_file_matching(self, path: str) -> Optional[str]:
        """Returns the path to a single file matching the given path including wildcards. Returns None if nothing
        found."""
        pass


class IFilesystemBehavior(IBehavior):
    @abc.abstractmethod
    def setup(self) -> IFilesystemSetup:
        pass

    @abc.abstractmethod
    def verify(self) -> IFilesystemVerify:
        pass

    @abc.abstractmethod
    def get_io_path(self) -> str:
        """Returns the path to the I/O files used by the service."""
        pass



import abc

from touchstone.lib.nodes.mocks.behaviors.i_behavior import IBehavior


class IHttpSetup(object):
    @abc.abstractmethod
    def get(self, endpoint: str, response: str, response_status: int = 200, url_pattern: bool = False,
            response_headers: dict = {'Content-Type': 'application/json'}):
        """Returns the given response when a GET request is made to the given endpoint."""
        pass

    @abc.abstractmethod
    def post(self, endpoint: str, response: str, response_status: int = 200, url_pattern: bool = False,
             response_headers: dict = {'Content-Type': 'application/json'}):
        """Returns the given response when a POST request is made to the given endpoint."""
        pass

    @abc.abstractmethod
    def put(self, endpoint: str, response: str, response_status: int = 200, url_pattern: bool = False,
            response_headers: dict = {'Content-Type': 'application/json'}):
        """Returns the given response when a PUT request is made to the given endpoint."""
        pass

    @abc.abstractmethod
    def delete(self, endpoint: str, response: str, response_status: int = 200, url_pattern: bool = False,
               response_headers: dict = {'Content-Type': 'application/json'}):
        """Returns the given response when a DELETE request is made to the given endpoint."""
        pass


class IHttpVerify(object):
    @abc.abstractmethod
    def get_called(self, endpoint: str, times: int = 1, url_pattern: bool = False) -> bool:
        """Returns True if the given endpoint has been called with a GET request the given number of times.
        If times is set to None, the endpoint can be called any number of times."""
        pass

    @abc.abstractmethod
    def post_called(self, endpoint: str, times: int = 1, url_pattern: bool = False) -> bool:
        """Returns True if the given endpoint has been called with a POST request the given number of times.
        If times is set to None, the endpoint can be called any number of times."""
        pass

    @abc.abstractmethod
    def post_contained(self, endpoint: str, expected_body: str, url_pattern: bool = False) -> bool:
        """Returns True if the given endpoint has been called with a POST request containing the given expected
        body."""
        pass

    @abc.abstractmethod
    def post_contained_json(self, endpoint: str, expected_body: dict, url_pattern: bool = False) -> bool:
        """Returns True if the given endpoint has been called with a POST request containing the given expected
        JSON body."""
        pass

    @abc.abstractmethod
    def put_called(self, endpoint: str, times: int = 1, url_pattern: bool = False) -> bool:
        """Returns True if the given endpoint has been called with a PUT request the given number of times.
        If times is set to None, the endpoint can be called any number of times."""
        pass

    @abc.abstractmethod
    def put_contained(self, endpoint: str, expected_body: str, url_pattern: bool = False) -> bool:
        """Returns True if the given endpoint has been called with a PUT request containing the given expected
        body."""
        pass

    @abc.abstractmethod
    def put_contained_json(self, endpoint: str, expected_body: dict, url_pattern: bool = False) -> bool:
        """Returns True if the given endpoint has been called with a PUT request containing the given expected
        JSON body."""
        pass

    @abc.abstractmethod
    def delete_called(self, endpoint: str, times: int = 1, url_pattern: bool = False) -> bool:
        """Returns True if the given endpoint has been called with a DELETE request the given number of times.
        If times is set to None, the endpoint can be called any number of times."""
        pass

    @abc.abstractmethod
    def delete_contained(self, endpoint: str, expected_body: str, url_pattern: bool = False) -> bool:
        """Returns True if the given endpoint has been called with a DELETE request containing the given expected
        body."""
        pass

    @abc.abstractmethod
    def delete_contained_json(self, endpoint: str, expected_body: dict, url_pattern: bool = False) -> bool:
        """Returns True if the given endpoint has been called with a DELETE request containing the given expected
        JSON body."""
        pass


class IHttpBehavior(IBehavior):
    @abc.abstractmethod
    def setup(self) -> IHttpSetup:
        pass

    @abc.abstractmethod
    def verify(self) -> IHttpVerify:
        pass

    @abc.abstractmethod
    def url(self) -> str:
        """Returns the URL of this mocked HTTP resource."""
        pass



import json
from json.decoder import JSONDecodeError

from pika.adapters.blocking_connection import BlockingChannel

from touchstone.helpers import validation
from touchstone.lib.nodes.mocks.behaviors.i_rabbitmq_behavior import IRabbitmqVerify
from touchstone.lib.nodes.mocks.docker.rabbitmq.docker_rabbitmq_context import DockerRabbitmqContext


class DockerRabbitmqVerify(IRabbitmqVerify):
    def __init__(self, rmq_context: DockerRabbitmqContext):
        super().__init__()
        self.__rmq_context = rmq_context
        self.__channel = None

    def set_blocking_channel(self, channel: BlockingChannel):
        self.__channel = channel

    def messages_published(self, exchange: str, num_expected: int = 1, routing_key: str = '') -> bool:
        if not self.__rmq_context.exchange_is_tracked(exchange, routing_key):
            return False
        num_messages = self.__rmq_context.messages_published(exchange, routing_key)
        if num_expected is None and num_messages != 0:
            return True
        return validation.matches(num_expected, num_messages)

    def payload_published(self, exchange: str, expected_payload: str, routing_key: str = '') -> bool:
        if not self.__rmq_context.exchange_is_tracked(exchange, routing_key):
            return False
        payloads = self.__rmq_context.payloads_published(exchange, routing_key)
        return validation.contains(expected_payload, payloads)

    def payload_published_json(self, exchange: str, expected_payload: dict, routing_key: str = '') -> bool:
        if not self.__rmq_context.exchange_is_tracked(exchange, routing_key):
            return False
        payloads = self.__rmq_context.payloads_published(exchange, routing_key)
        for payload in payloads:
            try:
                payload = json.loads(payload)
                if validation.matches(expected_payload, payload, quiet=True):
                    return True
            except JSONDecodeError:
                pass
        print(f'Expected:\n{expected_payload}\nwas not found in actual:\n{payloads}')
        return False



from typing import List

from touchstone.lib.listeners.i_services_available_listener import IServicesAvailableListener


class TsContext(object):
    def __init__(self):
        self.__services_available_listeners: List[IServicesAvailableListener] = []
        self.__triggered_services_available: bool = False

    def register_services_available_listener(self, listener: IServicesAvailableListener):
        self.__services_available_listeners.append(listener)

    def trigger_services_available(self):
        if not self.__triggered_services_available:
            self.__triggered_services_available = True
            for listener in self.__services_available_listeners:
                listener.services_available()



import logging

import click

from touchstone import common, __version__
from touchstone import develop
from touchstone import init
from touchstone import run
from touchstone.lib import exceptions

should_log_services: bool = False


@click.group()
@click.option('--log', default='WARNING', help='Sets the log level.')
@click.option('--log-services', is_flag=True, help="Captures service logs and stores them in 'touchstone/logs'.")
def cli(log, log_services):
    # Logging
    numeric_level = getattr(logging, log.upper(), None)
    if not isinstance(numeric_level, int):
        raise exceptions.TouchstoneException(f'Invalid log level: {log}')
    logging.basicConfig()
    common.logger.setLevel(numeric_level)

    # Log Services
    global should_log_services
    should_log_services = log_services


@cli.command('version', help='Prints the Touchstone version number.')
def cli_version():
    print(__version__)


@cli.command(name='init', help='Initialize Touchstone in the current directory.')
def init_cmd():
    init.execute()


@cli.command(name='run', help='Run all Touchstone tests and exit.')
def run_cmd():
    run.execute(should_log_services)


@cli.command(name='develop', help='Start a development session of Touchstone.')
def develop_cmd():
    develop.execute(should_log_services)


if __name__ == '__main__':
    cli()



from minio import Minio

from touchstone.lib import exceptions
from touchstone.lib.configurers.i_configurable import IConfigurable
from touchstone.lib.health_checks.http_health_check import HttpHealthCheck
from touchstone.lib.managers.docker_manager import DockerManager
from touchstone.lib.networking.docker_network import DockerNetwork
from touchstone.lib.networking.i_network import INetwork
from touchstone.lib.nodes.mocks.behaviors.i_behavior import IBehavior
from touchstone.lib.nodes.mocks.behaviors.i_s3_behavior import IS3Behavior, IS3Setup, IS3Verify
from touchstone.lib.nodes.mocks.docker.i_runnable_docker import IRunnableDocker
from touchstone.lib.nodes.mocks.docker.s3.docker_s3_setup import DockerS3Setup
from touchstone.lib.nodes.mocks.docker.s3.docker_s3_verify import DockerS3Verify


class DockerS3(IRunnableDocker, IS3Behavior):
    __USERNAME = 'admin123'
    __PASSWORD = 'admin123'

    def __init__(self, defaults_configurer: IConfigurable, base_objects_path: str, health_check: HttpHealthCheck,
                 setup: DockerS3Setup, verify: DockerS3Verify, docker_manager: DockerManager,
                 docker_network: DockerNetwork):
        self.__defaults_configurer = defaults_configurer
        self.__base_objects_path = base_objects_path
        self.__health_check = health_check
        self.__setup = setup
        self.__verify = verify
        self.__docker_manager = docker_manager
        self.__docker_network = docker_network

    def get_behavior(self) -> IBehavior:
        return self

    def get_network(self) -> INetwork:
        return self.__docker_network

    def initialize(self):
        s3_client = Minio(self.__docker_network.external_url(),
                          access_key=self.__USERNAME,
                          secret_key=self.__PASSWORD,
                          secure=False)
        self.__setup.set_s3_client(s3_client)
        self.__verify.set_s3_client(s3_client)
        self.__setup.init(self.__base_objects_path, self.__defaults_configurer.get_config())

    def start(self):
        run_result = self.__docker_manager.run_background_image('minio/minio:RELEASE.2020-02-27T00-23-05Z server /data',
                                                                port=9000,
                                                                environment_vars=[('MINIO_ACCESS_KEY', self.__USERNAME),
                                                                                  (
                                                                                      'MINIO_SECRET_KEY',
                                                                                      self.__PASSWORD)])
        self.__docker_network.set_container_id(run_result.container_id)
        self.__docker_network.set_internal_port(run_result.internal_port)
        self.__docker_network.set_external_port(run_result.external_port)
        self.__docker_network.set_ui_port(run_result.external_port)
        self.__docker_network.set_ui_endpoint('/minio')
        self.__docker_network.set_username(self.__USERNAME)
        self.__docker_network.set_password(self.__PASSWORD)

    def stop(self):
        if self.__docker_network.container_id():
            self.__docker_manager.stop_container(self.__docker_network.container_id())

    def reset(self):
        self.__setup.init(self.__base_objects_path, self.__defaults_configurer.get_config())

    def is_healthy(self) -> bool:
        self.__health_check.set_url(self.__docker_network.ui_url() + '/health/ready')
        return self.__health_check.is_healthy()

    def setup(self) -> IS3Setup:
        if not self.__setup:
            raise exceptions.MockException('Setup unavailable. Mock is still starting.')
        return self.__setup

    def verify(self) -> IS3Verify:
        if not self.__verify:
            raise exceptions.MockException('Verify unavailable. Mock is still starting.')
        return self.__verify

    def get_base_path(self) -> str:
        return self.__base_objects_path



import abc

from touchstone.lib.nodes.mocks.behaviors.i_behavior import IBehavior


class IRabbitmqSetup(object):
    @abc.abstractmethod
    def publish(self, exchange: str, payload: str, routing_key: str = '', content_type: str = None,
                headers: dict = None):
        """Publish a message with a payload to the given exchange and optional routing-key."""
        pass

    @abc.abstractmethod
    def publish_json(self, exchange: str, payload: dict, routing_key: str = '', headers: dict = None):
        """Publish a JSON message to the given exchange and optional routing-key. Content-type will be
        'application/json'."""
        pass


class IRabbitmqVerify(object):
    @abc.abstractmethod
    def messages_published(self, exchange: str, num_expected: int = 1, routing_key: str = '') -> bool:
        """Returns True if messages have been published the given number of times to the given exchange and
        routing-key. If num_expected is set to None, any number of messages will be considered passing."""
        pass

    @abc.abstractmethod
    def payload_published(self, exchange: str, expected_payload: str, routing_key: str = '') -> bool:
        """Returns True if a message with the given payload has been published to the given exchange and routing key."""
        pass

    @abc.abstractmethod
    def payload_published_json(self, exchange: str, expected_payload: dict, routing_key: str = '') -> bool:
        """Returns True if a JSON message with the given payload has been published to the given exchange and routing
        key."""
        pass


class IRabbitmqBehavior(IBehavior):
    DEFAULT_CONFIG = {
        'auto_create': True
    }

    @abc.abstractmethod
    def setup(self) -> IRabbitmqSetup:
        pass

    @abc.abstractmethod
    def verify(self) -> IRabbitmqVerify:
        pass



from touchstone import common
from touchstone.lib.configurers.i_configurable import IConfigurable


class BasicConfigurer(IConfigurable):
    def __init__(self, default_config: dict):
        self.__config = default_config

    def get_config(self) -> dict:
        return self.__config

    def merge_config(self, other: dict):
        self.__config = common.dict_merge(self.__config, other)



import json
from typing import List

from pymongo import MongoClient

from touchstone.lib.nodes.mocks.behaviors.i_mongodb_behavior import IMongodbSetup
from touchstone.lib.nodes.mocks.docker.mongodb.docker_mongo_context import DockerMongoContext


class DockerMongodbSetup(IMongodbSetup):
    def __init__(self, mongo_context: DockerMongoContext):
        self.__mongo_context = mongo_context
        self.__mongo_client = None

    def set_mongo_client(self, mongo_client: MongoClient):
        self.__mongo_client = mongo_client

    def init(self, defaults: dict):
        for database in self.__mongo_context.databases():
            mongo_database = self.__mongo_client.get_database(database)
            collections = mongo_database.list_collection_names(filter={"name": {"$regex": r"^(?!system\\.)"}})
            for collection in collections:
                mongo_database.drop_collection(collection)
        self.__mongo_context.clear()

        for database in defaults.get('databases', []):
            database_name = database['name']
            mongo_database = self.__mongo_client.get_database(database_name)
            self.__mongo_context.add_database(database_name)
            for collection in database.get('collections', []):
                mongo_collection = mongo_database.get_collection(collection['name'])
                self.__mongo_context.add_collection(database_name, collection['name'])
                for document in collection.get('documents', []):
                    # Supports multi-line documents
                    if isinstance(document, str):
                        document = json.loads(document)
                    mongo_collection.insert_one(document)

    def command(self, database: str, command: dict):
        if self.__mongo_context.database_exists(database):
            self.__mongo_client.get_database(database).command(command)

    def insert_document(self, database: str, collection: str, document: dict):
        if self.__mongo_context.collection_exists(database, collection):
            self.__mongo_client \
                .get_database(database) \
                .get_collection(collection).insert_one(document)

    def insert_documents(self, database: str, collection: str, documents: List[dict]):
        if self.__mongo_context.collection_exists(database, collection):
            self.__mongo_client \
                .get_database(database) \
                .get_collection(collection).insert_many(documents)



from pymongo import MongoClient

from touchstone.lib.nodes.mocks.behaviors.i_mongodb_behavior import IMongodbVerify
from touchstone.lib.nodes.mocks.docker.mongodb.docker_mongo_context import DockerMongoContext


class DockerMongodbVerify(IMongodbVerify):
    def __init__(self, mongo_context: DockerMongoContext):
        self.__mongo_context = mongo_context
        self.__mongo_client = None

    def set_mongo_client(self, mongo_client: MongoClient):
        self.__mongo_client = mongo_client

    def document_exists(self, database: str, collection: str, document: dict, num_expected: int = 1) -> bool:
        if not self.__mongo_context.collection_exists(database, collection):
            return False

        num_documents = self.__mongo_client \
            .get_database(database) \
            .get_collection(collection) \
            .count_documents(document)

        if num_expected is None and num_documents != 0:
            return True
        if num_expected == num_documents:
            return True
        print(f'Document: "{document}" in collection: "{collection}" was found {num_documents} time(s) but expected '
              f'{num_expected if num_expected else "any"} time(s).')
        return False



import abc
from typing import Tuple, List


class IRunnable(object):
    @abc.abstractmethod
    def start(self, environment_vars: List[Tuple[str, str]] = []):
        pass

    @abc.abstractmethod
    def stop(self):
        pass

    @abc.abstractmethod
    def is_running(self) -> bool:
        pass

    @abc.abstractmethod
    def wait_until_healthy(self):
        pass



from setuptools import setup, find_packages

from touchstone import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='touchstone-testing',
    version=__version__,
    description='Touchstone is a testing framework for your services that focuses on end-to-end and exploratory testing.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/shane-jansen/touchstone',
    author='Shane Jansen',
    author_email='shanejjansen@gmail.com',
    packages=find_packages(exclude=['tests']),
    install_requires=[
        'pyfiglet~=0.8',
        'click~=7.0',
        'pika~=1.1',
        'pyyaml~=5.4',
        'pymongo~=3.10',
        'pymysql~=0.9.3',
        'minio~=5.0.7',
        'redis~=3.5.3'
    ],
    entry_points={
        'console_scripts': [
            'touchstone = touchstone.cli:cli'
        ]
    },
    classifiers=[
        'Programming Language :: Python :: 3'
    ],
    python_requires='>=3.6'
)



import os
import re
import subprocess
import uuid
from typing import Optional, Tuple, List

from touchstone import common
from touchstone.lib import exceptions


class RunResult(object):
    def __init__(self, container_id, internal_port, external_port, ui_port):
        self.container_id = container_id
        self.internal_port = internal_port
        self.external_port = external_port
        self.ui_port = ui_port


class DockerManager(object):
    def __init__(self, should_auto_discover: bool = True):
        self.__images: list = []
        self.__containers: list = []
        self.__should_auto_discover = should_auto_discover
        self.__network: Optional[str] = None

    def build_dockerfile(self, dockerfile_path: str) -> str:
        # Build context will always be the same location as the Dockerfile for our purposes
        build_context = os.path.dirname(dockerfile_path)
        tag = uuid.uuid4().hex
        command = f'docker build -t {tag} -f {dockerfile_path} {build_context}'
        common.logger.debug(f'Building Dockerfile with command: {command}')
        result = subprocess.run(command, shell=True)
        if result.returncode is not 0:
            raise exceptions.ContainerException(f'An error occurred while building Dockerfile: "{dockerfile_path}".')
        self.__images.append(tag)
        return tag

    def run_background_image(self, image: str, port: int = None, exposed_port: int = None, ui_port: int = None,
                             environment_vars: List[Tuple[str, str]] = [], hostname: str = None,
                             log_path: str = None, options: str = None) -> RunResult:
        exposed_port = port if not exposed_port else exposed_port
        self.__create_network()

        additional_params = self.__build_ports_str(port, exposed_port, ui_port)
        if len(environment_vars) != 0:
            additional_params += ' '
            additional_params += self.__build_env_str(environment_vars)
        if options:
            additional_params += ' '
            additional_params += options

        container_id = self.__run_image(additional_params, image, log_path, hostname)

        # Extract the auto-discovered ports
        exposed_port = self.__extract_port_mapping(container_id, port)
        ui_port = self.__extract_port_mapping(container_id, ui_port)

        self.__containers.append(container_id)
        return RunResult(container_id, port, exposed_port, ui_port)

    def run_foreground_image(self, image: str, bind_mount: str, environment_vars: List[Tuple[str, str]] = [],
                             log_path: str = None, options: str = None):
        self.__create_network()
        additional_params = f'-v {bind_mount}'
        if len(environment_vars) != 0:
            additional_params += ' '
            additional_params += self.__build_env_str(environment_vars)
        if options:
            additional_params += ' '
            additional_params += options
        self.__execute_image(image, additional_params, log_path)

    def __create_network(self):
        if not self.__network:
            self.__network = uuid.uuid4().hex
            common.logger.debug(f'Creating network: {self.__network}')
            subprocess.run(['docker', 'network', 'create', self.__network], stdout=subprocess.DEVNULL)

    def __build_ports_str(self, port: int, exposed_port: int, ui_port: int) -> str:
        additional_params = ''
        if port:
            if self.__should_auto_discover:
                additional_params += f' -p :{port}'
            else:
                additional_params += f' -p {exposed_port}:{port}'
            additional_params += f' --expose {port}'
        if ui_port:
            additional_params += f' -p :{ui_port}'
        return additional_params[1:]

    def __build_env_str(self, environment_vars: List[Tuple[str, str]] = []) -> str:
        additional_params = ''
        for var, value in environment_vars:
            additional_params += f' -e {var}="{value}"'
        return additional_params[1:]

    def __run_image(self, additional_params: str, image: str, log_path: Optional[str], hostname: str = None) -> str:
        container_id = uuid.uuid4().hex
        command = f'docker run -d --network {self.__network} '
        if hostname:
            command += f'--hostname {hostname} '
        command += f'--name {container_id} {additional_params} {image}'
        common.logger.debug(f'Running container with command: {command}')
        result = subprocess.run(command, shell=True, stdout=subprocess.DEVNULL)
        if log_path:
            common.logger.debug(f'Writing container logs: {log_path}')
            with open(log_path, 'w') as file:
                subprocess.Popen(['docker', 'container', 'logs', '--follow', container_id], stdout=file)
        if result.returncode is not 0:
            raise exceptions.ContainerException(
                f'Container image {image} could not be started. Ensure Docker is running and ports are not already in '
                f'use.')
        return container_id

    def __execute_image(self, image: str, additional_params: str, log_path: Optional[str]):
        command = f'docker run --rm --network {self.__network} '
        command += f'{additional_params} {image}'
        common.logger.debug(f'Executing container with command: {command}')
        if log_path:
            with open(log_path, 'w') as file:
                result = subprocess.run(command, shell=True, stdout=file)
        else:
            result = subprocess.run(command, shell=True)
        if result.returncode is not 0:
            raise exceptions.ContainerException(f'Container finished execution with non-zero return code.')

    def __extract_port_mapping(self, container_id: str, given_port: int) -> int:
        result = str(subprocess.run(['docker', 'port', container_id], stdout=subprocess.PIPE).stdout,
                     encoding='utf-8')
        for line in result.splitlines():
            found_port = int(re.search('.+?(?=/)', line).group())
            new_port = re.search('(?<=0.0.0.0:)\\d*', line)
            if new_port:
                new_port = int(new_port.group())
            if found_port == given_port:
                return new_port
        return given_port

    def stop_container(self, id: str):
        common.logger.debug(f'Stopping container: {id}')
        subprocess.run(['docker', 'container', 'stop', id], stdout=subprocess.DEVNULL)
        subprocess.run(['docker', 'container', 'rm', '-v', id], stdout=subprocess.DEVNULL)
        self.__containers.remove(id)

    def cleanup(self):
        if self.__images:
            for image in self.__images:
                common.logger.debug(f'Removing image: {image}')
                subprocess.run(['docker', 'image', 'rm', image], stdout=subprocess.DEVNULL)
            self.__images = []
        if self.__containers:
            for container in self.__containers:
                common.logger.debug(f'Stopping container: {container}')
                subprocess.run(['docker', 'container', 'stop', container], stdout=subprocess.DEVNULL)
            self.__containers = []
        if self.__network:
            subprocess.run(['docker', 'network', 'rm', self.__network], stdout=subprocess.DEVNULL)
            self.__network = None

    def containers_running(self) -> bool:
        return len(self.__containers) > 0



import abc

from touchstone.lib.nodes.mocks.behaviors.i_behavior import IBehavior


class IMockable(object):
    @abc.abstractmethod
    def get_name(self) -> str:
        pass

    @abc.abstractmethod
    def get_pretty_name(self) -> str:
        pass

    @abc.abstractmethod
    def get_behavior(self) -> IBehavior:
        pass



import pymongo

from touchstone.lib import exceptions
from touchstone.lib.configurers.i_configurable import IConfigurable
from touchstone.lib.managers.docker_manager import DockerManager
from touchstone.lib.networking.docker_network import DockerNetwork
from touchstone.lib.networking.i_network import INetwork
from touchstone.lib.nodes.mocks.behaviors.i_behavior import IBehavior
from touchstone.lib.nodes.mocks.behaviors.i_mongodb_behavior import IMongodbBehavior, IMongodbSetup, IMongodbVerify
from touchstone.lib.nodes.mocks.docker.i_runnable_docker import IRunnableDocker
from touchstone.lib.nodes.mocks.docker.mongodb.docker_mongodb_setup import DockerMongodbSetup
from touchstone.lib.nodes.mocks.docker.mongodb.docker_mongodb_verify import DockerMongodbVerify


class DockerMongodb(IRunnableDocker, IMongodbBehavior):
    def __init__(self, defaults_configurer: IConfigurable, is_dev_mode: bool, setup: DockerMongodbSetup,
                 verify: DockerMongodbVerify, docker_manager: DockerManager, docker_network: DockerNetwork):
        self.__defaults_configurer = defaults_configurer
        self.__is_dev_mode = is_dev_mode
        self.__setup = setup
        self.__verify = verify
        self.__docker_manager = docker_manager
        self.__docker_network = docker_network
        self.__ui_container_id = None

    def get_behavior(self) -> IBehavior:
        return self

    def get_network(self) -> INetwork:
        return self.__docker_network

    def initialize(self):
        mongo_client = pymongo.MongoClient(self.__docker_network.external_host(), self.__docker_network.external_port())
        self.__setup.set_mongo_client(mongo_client)
        self.__verify.set_mongo_client(mongo_client)
        self.__setup.init(self.__defaults_configurer.get_config())

    def start(self):
        run_result = self.__docker_manager.run_background_image('mongo:4.0.14', port=27017)
        self.__docker_network.set_container_id(run_result.container_id)
        if self.__is_dev_mode:
            ui_run_result = self.__docker_manager.run_background_image('mongo-express:0.49.0',
                                                                       ui_port=8081,
                                                                       environment_vars=[
                                                                           ('ME_CONFIG_MONGODB_PORT',
                                                                            run_result.external_port),
                                                                           ('ME_CONFIG_MONGODB_SERVER',
                                                                            self.__docker_network.internal_host())])
            self.__ui_container_id = ui_run_result.container_id
            self.__docker_network.set_ui_port(ui_run_result.ui_port)
        self.__docker_network.set_internal_port(run_result.internal_port)
        self.__docker_network.set_external_port(run_result.external_port)

    def stop(self):
        if self.__docker_network.container_id():
            self.__docker_manager.stop_container(self.__docker_network.container_id())
        if self.__ui_container_id:
            self.__docker_manager.stop_container(self.__ui_container_id)

    def reset(self):
        self.__setup.init(self.__defaults_configurer.get_config())

    def is_healthy(self) -> bool:
        try:
            client = pymongo.MongoClient(self.__docker_network.external_host(), self.__docker_network.external_port())
            status = client.admin.command('serverStatus')['ok']
            return status == 1.0
        except Exception:
            return False

    def setup(self) -> IMongodbSetup:
        if not self.__setup:
            raise exceptions.MockException('Setup unavailable. Mock is still starting.')
        return self.__setup

    def verify(self) -> IMongodbVerify:
        if not self.__verify:
            raise exceptions.MockException('Verify unavailable. Mock is still starting.')
        return self.__verify



import abc

from touchstone.lib.nodes.mocks.behaviors.i_behavior import IBehavior


class IS3Setup(object):
    @abc.abstractmethod
    def create_bucket(self, name: str):
        """Creates a new bucket with the given name."""
        pass

    @abc.abstractmethod
    def put_object(self, bucket_name: str, object_name: str, data: bytes,
                   content_type: str = 'application/octet-stream'):
        """Puts the given bytes into a bucket and object."""
        pass

    @abc.abstractmethod
    def put_json_object(self, bucket_name: str, object_name: str, input_json: dict):
        """Puts the given JSON into a bucket and object."""
        pass


class IS3Verify(object):
    @abc.abstractmethod
    def bucket_exists(self, name: str) -> bool:
        """Return True if the given bucket name exists."""
        pass

    @abc.abstractmethod
    def object_exists(self, bucket_name: str, object_name: str) -> bool:
        """Returns True if the given bucket and object name exists."""
        pass

    @abc.abstractmethod
    def object_matches(self, bucket_name: str, object_name: str, expected: bytes) -> bool:
        """Returns True if the given object in a bucket matches the expected bytes."""
        pass

    @abc.abstractmethod
    def object_matches_json(self, bucket_name: str, object_name: str, expected_json: dict) -> bool:
        """Returns True if the given JSON object in a bucket matches the expected JSON."""
        pass


class IS3Behavior(IBehavior):
    @abc.abstractmethod
    def setup(self) -> IS3Setup:
        pass

    @abc.abstractmethod
    def verify(self) -> IS3Verify:
        pass

    @abc.abstractmethod
    def get_base_path(self) -> str:
        """Returns the "defaults" base directory path."""
        pass



import re

from pyfiglet import figlet_format

from touchstone import common
from touchstone.bootstrap import Bootstrap


def execute(should_log_services=False):
    if not common.sanity_check_passes():
        exit(1)
    bootstrap = Bootstrap(True, should_log_services)
    print(figlet_format('Touchstone', font='larry3d'))

    try:
        bootstrap.mocks.start()
        bootstrap.mocks.print_available_mocks()

        __print_help()
        while True:
            command = input('TS > ')
            if command == 'help':
                __print_help()
            elif command == 'run':
                bootstrap.ts_context.trigger_services_available()
                tests_did_pass = bootstrap.services.run_all_tests()
                if tests_did_pass:
                    print('All Touchstone tests passed successfully!')
                else:
                    print('One or more Touchstone tests failed.')
            elif re.search('run \\S* \\S* \\S*', command):
                command = command.lower()
                parts = command.split(' ')
                bootstrap.ts_context.trigger_services_available()
                bootstrap.services.run_test(parts[1], parts[2], parts[3])
            elif command == 'services start':
                try:
                    bootstrap.services.add_environment_vars(bootstrap.mocks.environment_vars())
                    bootstrap.services.start()
                    bootstrap.ts_context.trigger_services_available()
                except KeyboardInterrupt:
                    bootstrap.services.stop()
            elif command == 'services stop':
                bootstrap.services.stop()
            elif command == 'mocks print':
                bootstrap.mocks.print_available_mocks()
            elif command == 'mocks reset':
                bootstrap.mocks.reset()
            elif command == 'exit':
                bootstrap.exit(True)
            elif command == '':
                pass
            else:
                print(f'Unknown Touchstone command "{command}"')
    except (Exception, KeyboardInterrupt) as e:
        print('\nTouchstone was interrupted. Cleaning up...')
        bootstrap.cleanup()
        raise e


def __print_help():
    print('\nDevelopment mode:\n'
          'help - Prints this message.\n'
          'run - Runs all Touchstone tests.\n'
          'run {service} {test file name} {test name} - Run a single test. Leaves mocks in the tested state.\n'
          'services start - If you don\'t want to run services manually, this starts all networked services with a non-empty Dockerfile.\n'
          'services stop - Stops all started containerized services.\n'
          'mocks print - Prints mock UI URLs.\n'
          'mocks reset - Resets all mocks to their default state.\n'
          'exit - Exit Touchstone.\n')



import json
from threading import Thread

import pika
import pika.exceptions
from pika import spec, BasicProperties
from pika.adapters.blocking_connection import BlockingChannel

from touchstone.lib.nodes.mocks.behaviors.i_rabbitmq_behavior import IRabbitmqSetup
from touchstone.lib.nodes.mocks.docker.rabbitmq.docker_rabbitmq_context import DockerRabbitmqContext


class MessageConsumer(Thread):
    def __init__(self, connection_params: pika.ConnectionParameters, rmq_context: DockerRabbitmqContext):
        super().__init__()
        self.__connection_params = connection_params
        self.__rmq_context = rmq_context

        connection = pika.BlockingConnection(self.__connection_params)
        self.channel: BlockingChannel = connection.channel()

    def run(self) -> None:
        super().run()
        try:
            self.channel.start_consuming()
        except pika.exceptions.ConnectionClosedByBroker:
            pass

    def consume(self, exchange: str, routing_key: str, queue: str):
        def message_received(channel: BlockingChannel, method: spec.Basic.Deliver, properties: spec.BasicProperties,
                             body: bytes):
            payload = str(body, encoding='utf-8')
            self.__rmq_context.track_payload(exchange, routing_key, payload)
            channel.basic_ack(delivery_tag=method.delivery_tag)

        self.channel.basic_consume(queue, message_received)


class DockerRabbitmqSetup(IRabbitmqSetup):
    def __init__(self, rmq_context: DockerRabbitmqContext):
        super().__init__()
        self.__rmq_context = rmq_context
        self.__channel = None
        self.__message_consumer = None
        self.__exchanges: list = []
        self.__queues: list = []
        self.__shadow_queues: list = []

    def set_channel(self, channel: BlockingChannel):
        self.__channel = channel

    def set_connection_params(self, connection_params: pika.ConnectionParameters):
        self.__message_consumer = MessageConsumer(connection_params, self.__rmq_context)

    def purge_queues(self):
        for queue in self.__queues:
            self.__channel.queue_purge(queue)
        for queue in self.__shadow_queues:
            self.__channel.queue_purge(queue)
        self.__rmq_context.clear()

    def create_all(self, defaults: dict):
        for exchange in defaults.get('exchanges', []):
            self.__create_exchange(exchange['name'], exchange['type'])
            for queue in exchange.get('queues', []):
                routing_key = queue.get('routingKey', '')
                self.__create_queue(queue['name'], exchange['name'], routing_key)
                self.__create_shadow_queue(queue['name'], exchange['name'], routing_key)
        self.__start_listener()

    def create_shadow_queues(self, defaults: dict):
        for exchange in defaults.get('exchanges', []):
            for queue in exchange.get('queues', []):
                routing_key = queue.get('routingKey', '')
                self.__create_shadow_queue(queue['name'], exchange['name'], routing_key)
        self.__start_listener()

    def stop_listening(self):
        def callback():
            self.__message_consumer.channel.stop_consuming()

        if self.__message_consumer and self.__message_consumer.is_alive():
            self.__message_consumer.channel.connection.add_callback_threadsafe(callback)
            self.__message_consumer.join()

    def publish(self, exchange: str, payload: str, routing_key: str = '', content_type: str = None,
                headers: dict = None):
        if self.__rmq_context.exchange_is_tracked(exchange, routing_key):
            properties = BasicProperties(content_type=content_type, headers=headers)
            self.__channel.basic_publish(exchange, routing_key, bytes(payload, encoding='utf-8'), properties=properties)

    def publish_json(self, exchange: str, payload: dict, routing_key: str = '', headers: dict = None):
        self.publish(exchange, json.dumps(payload), routing_key, 'application/json', headers)

    def __start_listener(self):
        if not self.__message_consumer.is_alive() and (self.__queues or self.__shadow_queues):
            self.__message_consumer.start()

    def __create_exchange(self, name: str, exchange_type: str = 'direct'):
        if name not in self.__exchanges:
            self.__channel.exchange_declare(name, exchange_type=exchange_type)
            self.__exchanges.append(name)

    def __create_queue(self, name: str, exchange: str, routing_key: str = ''):
        if name not in self.__queues:
            self.__channel.queue_declare(name)
            self.__channel.queue_bind(name, exchange, routing_key=routing_key)
            self.__queues.append(name)

    def __create_shadow_queue(self, name: str, exchange: str, routing_key: str = ''):
        name = f'{name}-{routing_key}.ts-shadow'
        if name not in self.__shadow_queues:
            self.__channel.queue_declare(name)
            self.__channel.queue_bind(name, exchange, routing_key=routing_key)
            self.__shadow_queues.append(name)
            self.__rmq_context.track(exchange, routing_key)
            self.__message_consumer.consume(exchange, routing_key, name)



from pymysql.cursors import Cursor

from touchstone import common
from touchstone.helpers import validation
from touchstone.lib.nodes.mocks.behaviors.i_mysql_behabior import IMysqlVerify
from touchstone.lib.nodes.mocks.docker.mysql.docker_mysql_context import DockerMysqlContext


class DockerMysqlVerify(IMysqlVerify):
    def __init__(self, mysql_context: DockerMysqlContext):
        self.__mysql_context = mysql_context
        self.__cursor = None
        self.__convert_camel_to_snake = False

    def set_cursor(self, cursor: Cursor):
        self.__cursor = cursor

    def set_convert_camel_to_snake(self, convert_camel_to_snake: bool):
        self.__convert_camel_to_snake = convert_camel_to_snake

    def row_exists(self, database: str, table: str, where_conditions: dict, num_expected: int = 1) -> bool:
        if not self.__mysql_context.database_exists(database):
            return False

        if self.__convert_camel_to_snake:
            where_conditions = common.to_snake(where_conditions)
        where = []
        user_values = {}
        for key, value in where_conditions.items():
            if value is None:
                where.append(f'{key} IS NULL')
            elif value is validation.ANY:
                where.append(f'{key} IS NOT NULL')
            else:
                where.append(f'{key}=%({key})s')
                user_values[key] = value
        sql = f"SELECT COUNT(*) FROM {table} WHERE {' AND '.join(where)}"
        common.logger.debug(f'Executing: {sql}')
        self.__cursor.execute(f'USE {database}')
        self.__cursor.execute(sql, user_values)
        num_rows = self.__cursor.fetchone()['COUNT(*)']

        if num_expected is None and num_rows != 0:
            return True
        if num_expected == num_rows:
            return True
        print(f'SQL: "{sql}" in database: "{database}" was found {num_rows} time(s) but expected '
              f'{num_expected if num_expected else "any"} time(s).')
        return False

    def row_does_not_exist(self, database: str, table: str, where_conditions: dict) -> bool:
        return self.row_exists(database, table, where_conditions, num_expected=0)



import time

from touchstone.lib.health_checks.i_health_checkable import IHealthCheckable


class BlockingHealthCheck(object):
    def __init__(self, seconds_between_retries: int, num_retries: int, target: IHealthCheckable):
        self.__seconds_between_retries = seconds_between_retries
        self.__num_retries = num_retries
        self.__target = target

    def get_target(self) -> IHealthCheckable:
        return self.__target

    def wait_until_healthy(self) -> bool:
        for retry_num in range(self.__num_retries):
            if self.__target.is_healthy():
                return True
            else:
                time.sleep(self.__seconds_between_retries)
        return False
